{-
Copyright 2009 Christian Sternagel, René Thiemann, Sarah Winkler, Harald Zankl

This file is part of IsaFoR/CeTA.

IsaFoR/CeTA is free software: you can redistribute it and/or modify it under the
terms of the GNU Lesser General Public License as published by the Free Software
Foundation, either version 3 of the License, or (at your option) any later
version.

IsaFoR/CeTA is distributed in the hope that it will be useful, but WITHOUT ANY
WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR A
PARTICULAR PURPOSE.  See the GNU Lesser General Public License for more details.

You should have received a copy of the GNU Lesser General Public License along
with IsaFoR/CeTA. If not, see <http://www.gnu.org/licenses/>.
-}
module Term where {

import qualified List;
import qualified Show;

data Term a b = Fun a [(Term.Term a b)] | Var b deriving (Read, Show);

sshow_term ::
  forall c d. (Show.Sshow c, Show.Sshow d) => Term.Term c d -> [Char];
sshow_term (Term.Fun f ts) =
  List.append (Show.sshow f)
    (case ts of {
      [] -> Show.sEmpty;
      s : ss ->
        List.append "("
          (List.append
            (List.append (Term.sshow_term s)
              (List.concata
                (List.mapa (\ x -> List.concata [",", Term.sshow_term x]) ss)))
            ")");
    });
sshow_term (Term.Var x) = Show.sshow x;

isVar :: forall f v. Term.Term f v -> Bool;
isVar (Term.Fun uu uv) = False;
isVar (Term.Var x) = True;

eq_term :: forall f v. (Eq f, Eq v) => Term.Term f v -> Term.Term f v -> Bool;
eq_term (Term.Var v) (Term.Var v') = v == v';
eq_term (Term.Fun f list) (Term.Fun f' list') = f == f' && list == list';
eq_term (Term.Var a) (Term.Fun b c) = False;
eq_term (Term.Fun b c) (Term.Var a) = False;

instance (Eq a, Eq b) => Eq (Term.Term a b) where {
  (==) = Term.eq_term;
};

instance (Show.Sshow a, Show.Sshow b) => Show.Sshow (Term.Term a b) where {
  sshow = Term.sshow_term;
};

}
