(*
Copyright 2009 Christian Sternagel, René Thiemann, Sarah Winkler, Harald Zankl

This file is part of IsaFoR/CeTA.

IsaFoR/CeTA is free software: you can redistribute it and/or modify it under the
terms of the GNU Lesser General Public License as published by the Free Software
Foundation, either version 3 of the License, or (at your option) any later
version.

IsaFoR/CeTA is distributed in the hope that it will be useful, but WITHOUT ANY
WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR A
PARTICULAR PURPOSE.  See the GNU Lesser General Public License for more details.

You should have received a copy of the GNU Lesser General Public License along
with IsaFoR/CeTA. If not, see <http://www.gnu.org/licenses/>.
*)
theory AbstractRewriting imports Relation_Power Hilbert_Choice begin

section {* Abstract Rewrite Systems *}
types 
 'a brel = "('a \<times> 'a)set"

subsection {* Definitions *}
definition joinable :: "'a brel \<Rightarrow> 'a brel" ("(_^\<down>)" [1000] 999) where
 "r^\<down> \<equiv> ((r\<inverse>)^* O r^*)"

definition meetable :: "'a brel \<Rightarrow> 'a brel" ("(_^\<up>)" [1000] 999) where
 "r^\<up> \<equiv> (r^* O (r\<inverse>)^*)"

abbreviation symmc :: "'a brel \<Rightarrow> 'a brel" ("(_^\<leftrightarrow>)" [1000] 999) where
 "r^\<leftrightarrow> \<equiv> r \<union> r\<inverse>"

definition convertible :: "'a brel \<Rightarrow> 'a brel" ("(_^\<leftrightarrow>*)" [1000] 999) where
 "r^\<leftrightarrow>* \<equiv> (r^\<leftrightarrow>)^*"

inductive_set NF :: "'a brel \<Rightarrow> 'a set"
for r :: "'a brel" where
 intro[intro]: "r``{x} = {} \<Longrightarrow> x \<in> NF r"

lemma joinable_I[intro]: "(x,y) \<in> r^* \<Longrightarrow> (z,y) \<in> r^* \<Longrightarrow> (x,z) \<in> r^\<down>"
unfolding joinable_def rtrancl_converse by auto

lemma joinable_D: "(x,z) \<in> r^\<down> \<Longrightarrow> \<exists>y. (x,y) \<in> r^* \<and> (z,y) \<in> r^*"
unfolding joinable_def rtrancl_converse by best

lemma joinable_E[consumes 1,elim]:
 assumes "(x,z) \<in> r^\<down>" and "\<And>y. (x,y) \<in> r^* \<Longrightarrow> (z,y) \<in> r^* \<Longrightarrow> P"
 shows "P"
proof -
 from `(x,z) \<in> r^\<down>` obtain y where "(x,y) \<in> r^*" and "(z,y) \<in> r^*" using joinable_D by best
 with assms show ?thesis by simp
qed

lemma meetable_I[intro]: "(y,x) \<in> r^* \<Longrightarrow> (y,z) \<in> r^* \<Longrightarrow> (x,z) \<in> r^\<up>"
unfolding meetable_def rtrancl_converse by auto

lemma meetable_D: "(x,z) \<in> r^\<up> \<Longrightarrow> \<exists>y. (y,x) \<in> r^* \<and> (y,z) \<in> r^*"
unfolding meetable_def rtrancl_converse by best

lemma meetable_E[consumes 1,elim]:
 assumes "(y,z) \<in> r^\<up>" and "\<And>x. (x,y) \<in> r^* \<Longrightarrow> (x,z) \<in> r^* \<Longrightarrow> P"
 shows "P"
proof -
 from `(y,z) \<in> r^\<up>` obtain x where "(x,y) \<in> r^*" and "(x,z) \<in> r^*" using meetable_D by best
 with assms show ?thesis by simp
qed

lemma symmc_I1[intro]: "(x,y) \<in> r \<Longrightarrow> (x,y) \<in> r^\<leftrightarrow>" by auto

lemma symmc_I2[intro]: "(y,x) \<in> r \<Longrightarrow> (x,y) \<in> r^\<leftrightarrow>" by auto

lemma symmc_E[elim]: "(x,y) \<in> r^\<leftrightarrow> \<Longrightarrow> ((x,y) \<in> r \<Longrightarrow> P) \<Longrightarrow> ((y,x) \<in> r \<Longrightarrow> P) \<Longrightarrow> P" by auto

lemma symmc_D: "(x,y) \<in> r^\<leftrightarrow> \<Longrightarrow> (x,y) \<in> r \<or> (y,x) \<in> r" by simp

lemma convertible_I[intro]: "(x,y) \<in> r^* \<Longrightarrow> (x,y) \<in> r^\<leftrightarrow>*" unfolding convertible_def
proof (induct rule: rtrancl_induct)
 case base thus ?case by simp
next
 case (step y z)
 hence "(y,z) \<in> r^\<leftrightarrow>" by simp
 with `(x,y) \<in> (r^\<leftrightarrow>)^*` show ?case by (rule rtrancl.intros(2))
qed

lemma convertible_I2[intro]: "(x,y) \<in> (r^\<leftrightarrow>)^* \<Longrightarrow> (x,y) \<in> r^\<leftrightarrow>*" unfolding convertible_def by simp

lemma convertible_E[elim]: "(x,y) \<in> r^\<leftrightarrow>* \<Longrightarrow> ((x,y) \<in> (r^\<leftrightarrow>)^* \<Longrightarrow> P) \<Longrightarrow> P" unfolding convertible_def by simp

lemma convertible_trans: "trans(r^\<leftrightarrow>*)" unfolding trans_def
proof (intro allI impI)
 fix x y z assume "(x,y) \<in> r^\<leftrightarrow>*" and "(y,z) \<in> r^\<leftrightarrow>*" thus "(x,z) \<in> r^\<leftrightarrow>*" unfolding convertible_def
 proof induct
  case base thus ?case by simp
 next
  case (step y z')
  from `(y,z') \<in> r^\<leftrightarrow>` and `(z',z) \<in> (r^\<leftrightarrow>)^*` have "(y,z) \<in> (r^\<leftrightarrow>)^*" by (rule converse_rtrancl_into_rtrancl)
  with step show ?case by simp
 qed
qed

lemma convertible_sym: "sym(r^\<leftrightarrow>*)" unfolding sym_def
proof (intro allI impI)
 fix x y assume "(x,y) \<in> r^\<leftrightarrow>*" thus "(y,x) \<in> r^\<leftrightarrow>*" unfolding convertible_def
 proof (induct)
  case base thus ?case by simp
 next
  case (step y z)
  hence "(z,y) \<in> r^\<leftrightarrow>" by auto
  from `(z,y) \<in> r^\<leftrightarrow>` and `(y,x) \<in> (r^\<leftrightarrow>)^*` show ?case by (rule converse_rtrancl_into_rtrancl)
 qed
qed

lemma joinable_left_I[intro]:
 assumes "(x,y) \<in> r^*" and "(y,z) \<in> r^\<down>" shows "(x,z) \<in> r^\<down>"
proof -
 from `(y,z) \<in> r^\<down>` obtain y' where "(y,y') \<in> r^*" and "(y',z) \<in> (r\<inverse>)^*" unfolding joinable_def by best
 with `(x,y) \<in> r^*` have "(x,y') \<in> r^*" by simp
 with `(y',z) \<in> (r\<inverse>)^*` show ?thesis unfolding joinable_def by auto
qed

lemma joinable_right_I[intro]:
 assumes "(x,y) \<in> r^\<down>" and "(z,y) \<in> r^*" shows "(x,z) \<in> r^\<down>"
proof -
 from `(z,y) \<in> r^*` have "(y,z) \<in> (r\<inverse>)^*" unfolding rtrancl_converse by simp
 from `(x,y) \<in> r^\<down>` obtain x' where "(x,x') \<in> r^*" and "(x',y) \<in> (r\<inverse>)^*" unfolding joinable_def by best
 with `(y,z) \<in> (r\<inverse>)^*` have "(x',z) \<in> (r\<inverse>)^*" by simp
 with `(x,x') \<in> r^*` show ?thesis unfolding joinable_def by auto
qed

lemma NF_I[intro]: "\<forall>y. (x,y) \<notin> r \<Longrightarrow> x \<in> NF r" by auto

lemma NF_E[consumes 1,elim]: "x \<in> NF r \<Longrightarrow> (r``{x} = {} \<Longrightarrow> P) \<Longrightarrow> P" by (rule NF.cases) auto

lemma NF_no_step: "x \<in> NF r \<Longrightarrow> \<forall>y. (x,y) \<notin> r" by auto

lemma NF_E2[elim]: "x \<in> NF r \<Longrightarrow> ((!!y. (x,y) \<notin> r) \<Longrightarrow> P) \<Longrightarrow> P" by auto

lemma NF_no_trancl_step: assumes "x \<in> NF r" shows "\<forall>y. (x,y) \<notin> r^+"
proof -
 from assms have notin_r:"\<forall>y. (x,y) \<notin> r" by auto
 show ?thesis
 proof (intro allI notI)
  fix y assume "(x,y) \<in> r^+" thus False by induct (auto simp:notin_r)
 qed
qed

definition normalizable :: "'a brel \<Rightarrow> 'a brel" ("(_^!)" [1000] 999) where
 "r^! \<equiv> {(x,y). (x,y) \<in> r^* \<and> y \<in> NF r}"

lemma normalizable_I[intro]: "(x,y) \<in> r^* \<Longrightarrow> y \<in> NF r \<Longrightarrow> (x,y) \<in> r^!" unfolding normalizable_def by auto

lemma normalizable_I2[intro]: "(x,y) \<in> r^* \<Longrightarrow> (y,z) \<in> r^! \<Longrightarrow> (x,z) \<in> r^!" unfolding normalizable_def by auto

lemma normalizable_E[elim]: "(x,y) \<in> r^! \<Longrightarrow> ((x,y) \<in> r^* \<Longrightarrow> y \<in> NF r \<Longrightarrow> P) \<Longrightarrow> P"
unfolding normalizable_def by auto

subsubsection {* Exercises *}
text {* Aart's Lecture Notes (ALN), Exercise~1.6.a. *}
lemma NF_trancl_iff_NF:"NF(r^+) = NF r" by (auto simp:Image_mono NF_no_trancl_step)

text {* ALN, Exercise~1.6.b. *}
lemma NF_subset: "r \<subseteq> s \<Longrightarrow> NF s \<subseteq> NF r" by auto

text {* ALN, Exercise~1.6.c. *}
lemma joinability_refl':"(x,x) \<in> r^\<down>" unfolding joinable_def by auto
lemma joinability_refl:"reflexive(r^\<down>)" unfolding refl_def using joinability_refl' by auto
lemma refl_imp_NF_empty: "reflexive(r) \<Longrightarrow> NF(r) = {}" unfolding refl_def by auto
lemma NF_joinability:"NF(r^\<down>) = {}" by (rule refl_imp_NF_empty[OF joinability_refl])

subsection {* Properties *}
text {* complete, Church-Rosser property, semi-complete, strong normalization, unique normal forms, 
weak Church-Rosser property, weak normalization *}
consts
 complete :: "'a \<Rightarrow> 'b"
 CR  :: "'a \<Rightarrow> 'b"
 semi_complete :: "'a \<Rightarrow> 'b"
 SN  :: "'a \<Rightarrow> 'b"
 UNF :: "'a \<Rightarrow> 'b"
 WCR :: "'a \<Rightarrow> 'b"
 WN  :: "'a \<Rightarrow> 'b"


defs (overloaded) 
 CR_elt_def: "(CR::'a brel \<Rightarrow> 'a set) r x \<equiv> (\<forall>y z. (x,y) \<in> r^* \<and> (x,z) \<in> r^* \<longrightarrow> (y,z) \<in> r^\<down>)"
 CR_ars_def: "(CR::'a brel \<Rightarrow> bool) r \<equiv> (\<forall>(x::'a). CR r x)"
 SN_elt_def: "(SN::'a brel \<Rightarrow> 'a set) r x \<equiv> \<not>(\<exists>S. S 0 = x \<and> (\<forall>i. (S i,S(Suc i)) \<in> r))"
 SN_ars_def: "(SN::'a brel \<Rightarrow> bool) r \<equiv> (\<forall>(x::'a). SN r x)"
 UNF_elt_def: "(UNF::'a brel \<Rightarrow> 'a set) r x \<equiv> (\<forall>y z. (x,y) \<in> r^! \<and> (x,z) \<in> r^! \<longrightarrow> y = z)"
 UNF_ars_def: "(UNF::'a brel \<Rightarrow> bool) r \<equiv> (\<forall>(x::'a). UNF r x)"
 WCR_elt_def: "(WCR::'a brel \<Rightarrow> 'a set) r x \<equiv> (\<forall>y z. (x,y) \<in> r \<and> (x,z) \<in> r \<longrightarrow> (y,z) \<in> r^\<down>)"
 WCR_ars_def: "(WCR::'a brel \<Rightarrow> bool) r \<equiv> (\<forall>(x::'a). WCR r x)"
 WN_elt_def: "(WN::'a brel \<Rightarrow> 'a set) r x \<equiv> (\<exists>y. (x,y) \<in> r^!)"
 WN_ars_def: "(WN::'a brel \<Rightarrow> bool) r \<equiv> (\<forall>(x::'a). WN r x)"

defs (overloaded)
 complete_elt_def: "(complete::'a brel \<Rightarrow> 'a set) r x \<equiv> (SN r x \<and> CR r x)"
 complete_ars_def: "(complete::'a brel \<Rightarrow> bool) r \<equiv> (\<forall>(x::'a). complete r x)"
 semi_complete_elt_def: "(semi_complete::'a brel \<Rightarrow> 'a set) r x \<equiv> (WN r x \<and> CR r x)"
 semi_complete_ars_def: "(semi_complete::'a brel \<Rightarrow> bool) r \<equiv> (\<forall>(x::'a). semi_complete r x)"

text {* unique normal forms with respect to conversion *}
definition UNC :: "'a brel \<Rightarrow> bool" where
 "UNC r \<equiv> (\<forall>x y. x \<in> NF r \<and> y \<in> NF r \<and> (x,y) \<in> r^\<leftrightarrow>* \<longrightarrow> x = y)"

lemma complete_elt_I[intro]: "SN r x \<Longrightarrow> CR r x \<Longrightarrow> (complete::'a brel \<Rightarrow> 'a set) r x"
unfolding complete_elt_def by simp

lemma complete_elt_E[elim]: "(complete::'a brel \<Rightarrow> 'a set) r x \<Longrightarrow> (SN r x \<Longrightarrow> CR r x \<Longrightarrow> P) \<Longrightarrow> P"
unfolding complete_elt_def by simp

lemma complete_ars_I[intro]: "(!!(x::'a). complete r x) \<Longrightarrow> (complete::'a brel \<Rightarrow> bool) r"
unfolding complete_ars_def by simp

lemma complete_ars_I2[intro]:
 "(!!(x::'a). SN r x \<and> CR r x) \<Longrightarrow> (complete::'a brel \<Rightarrow> bool) r"
unfolding complete_ars_def complete_elt_def by simp

lemma complete_ars_E[elim]:
 "(complete::'a brel \<Rightarrow> bool) r \<Longrightarrow> ((!!(x::'a). complete r x) \<Longrightarrow> P) \<Longrightarrow> P"
unfolding complete_ars_def by simp

lemma complete_ars_E2[elim]:
 "(complete::'a brel \<Rightarrow> bool) r \<Longrightarrow> ((!!(x::'a). SN r x \<Longrightarrow> CR r x) \<Longrightarrow> P) \<Longrightarrow> P"
unfolding complete_ars_def complete_elt_def by simp

lemma compete_ars_E3[elim]:
 "(complete::'a brel \<Rightarrow> bool) r \<Longrightarrow> (SN r \<Longrightarrow> CR r \<Longrightarrow> P) \<Longrightarrow> P"
unfolding complete_ars_def complete_elt_def SN_ars_def CR_ars_def by simp

lemma CR_elt_I[intro]:
 "(!!y z. (x,y) \<in> r^* \<Longrightarrow> (x,z) \<in> r^* \<Longrightarrow> (y,z) \<in> r^\<down>) \<Longrightarrow> (CR::'a brel \<Rightarrow> 'a set) r x"
unfolding CR_elt_def by simp

lemma CR_elt_E[elim]:
 "(CR::'a brel \<Rightarrow> 'a set) r x \<Longrightarrow> ((!!y z. (x,y) \<in> r^* \<Longrightarrow> (x,z) \<in> r^* \<Longrightarrow> (y,z) \<in> r^\<down>) \<Longrightarrow> P) \<Longrightarrow> P"
unfolding CR_elt_def by simp

lemma CR_ars_I[intro]: "(!!(x::'a). CR r x) \<Longrightarrow> (CR::'a brel \<Rightarrow> bool) r"
unfolding CR_ars_def by simp

lemma CR_ars_I2[intro]:
 "(!!(x::'a) y z. (x,y) \<in> r^* \<Longrightarrow> (x,z) \<in> r^* \<Longrightarrow> (y,z) \<in> r^\<down>) \<Longrightarrow> (CR::'a brel \<Rightarrow> bool) r"
unfolding CR_ars_def CR_elt_def by auto

lemma CR_ars_E[elim]: "(CR::'a brel \<Rightarrow> bool) r \<Longrightarrow> ((!!(x::'a). CR r x) \<Longrightarrow> P) \<Longrightarrow> P"
unfolding CR_ars_def by simp

lemma CR_ars_E2[elim]:
 "(CR::'a brel \<Rightarrow> bool) r \<Longrightarrow> ((!!(x::'a) y z. (x,y) \<in> r^* \<Longrightarrow> (x,z) \<in> r^* \<Longrightarrow> (y,z) \<in> r^\<down>) \<Longrightarrow> P) \<Longrightarrow> P"
unfolding CR_ars_def CR_elt_def by auto

lemma semi_complete_elt_I[intro]: "WN r x \<Longrightarrow> CR r x \<Longrightarrow> (semi_complete::'a brel \<Rightarrow> 'a set) r x"
unfolding semi_complete_elt_def by simp

lemma semi_complete_elt_E[elim]:
 "(semi_complete::'a brel \<Rightarrow> 'a set) r x \<Longrightarrow> (WN r x \<Longrightarrow> CR r x \<Longrightarrow> P) \<Longrightarrow> P"
unfolding semi_complete_elt_def by simp

lemma semi_complete_ars_I[intro]:
 "(!!(x::'a). semi_complete r x) \<Longrightarrow> (semi_complete::'a brel \<Rightarrow> bool) r"
unfolding semi_complete_ars_def by simp

lemma semi_complete_ars_I2[intro]:
 "(!!(x::'a). WN r x \<and> CR r x) \<Longrightarrow> (semi_complete::'a brel \<Rightarrow> bool) r"
unfolding semi_complete_ars_def semi_complete_elt_def by auto

lemma semi_complete_ars_E[elim]:
 "(semi_complete::'a brel \<Rightarrow> bool) r \<Longrightarrow> ((!!(x::'a). semi_complete r x) \<Longrightarrow> P) \<Longrightarrow> P"
unfolding semi_complete_ars_def by simp

lemma semi_complete_ars_E2[elim]:
 "(semi_complete::'a brel \<Rightarrow> bool) r \<Longrightarrow> ((!!(x::'a). WN r x \<Longrightarrow> CR r x) \<Longrightarrow> P) \<Longrightarrow> P"
unfolding semi_complete_ars_def semi_complete_elt_def by simp

lemma semi_compete_ars_E3[elim]:
 "(semi_complete::'a brel \<Rightarrow> bool) r \<Longrightarrow> (WN r \<Longrightarrow> CR r \<Longrightarrow> P) \<Longrightarrow> P"
unfolding semi_complete_ars_def semi_complete_elt_def WN_ars_def CR_ars_def by simp

lemma UNC_I[intro]: "(!!x y. x \<in> NF r \<Longrightarrow> y \<in> NF r \<Longrightarrow> (x,y) \<in> r^\<leftrightarrow>* \<Longrightarrow> x = y) \<Longrightarrow> UNC r"
unfolding UNC_def by simp

lemma UNC_E[elim]: "UNC r \<Longrightarrow> ((!!x y. x \<in> NF r \<Longrightarrow> y \<in> NF r \<Longrightarrow> (x,y) \<in> r^\<leftrightarrow>* \<Longrightarrow> x = y) \<Longrightarrow> P) \<Longrightarrow> P"
unfolding UNC_def by simp

lemma UNF_elt_I[intro]: "(!!y z. (x,y) \<in> r^! \<Longrightarrow> (x,z) \<in> r^! \<Longrightarrow> y = z) \<Longrightarrow> (UNF::'a brel \<Rightarrow> 'a set) r x"
unfolding UNF_elt_def by simp

lemma UNF_elt_E[elim]:
 "(UNF::'a brel \<Rightarrow> 'a set) r x \<Longrightarrow> ((!!y z. (x,y) \<in> r^! \<Longrightarrow> (x,z) \<in> r^! \<Longrightarrow> y = z) \<Longrightarrow> P) \<Longrightarrow> P"
unfolding UNF_elt_def by simp

lemma UNF_ars_I[intro]: "(!!(x::'a). UNF r x) \<Longrightarrow> (UNF::'a brel \<Rightarrow> bool) r"
unfolding UNF_ars_def by simp

lemma UNF_ars_I2[intro]:
 "(!!(x::'a) y z. (x,y) \<in> r^! \<Longrightarrow> (x,z) \<in> r^! \<Longrightarrow> y = z) \<Longrightarrow> (UNF::'a brel \<Rightarrow> bool) r"
unfolding UNF_ars_def UNF_elt_def by auto

lemma SN_elt_I[intro]: "\<not>(\<exists>S. S 0 = x \<and> (\<forall>i. (S i,S(Suc i)) \<in> r)) \<Longrightarrow> SN r x"
unfolding SN_elt_def by simp

lemma SN_elt_E[elim]: "SN r x \<Longrightarrow> ((!!S. S 0 = x \<Longrightarrow> (\<exists>i. (S i,S(Suc i)) \<notin> r)) \<Longrightarrow> P) \<Longrightarrow> P"
unfolding SN_elt_def by simp

lemma SN_ars_I[intro]: "(!!(x::'a). SN r x) \<Longrightarrow> (SN::'a brel \<Rightarrow> bool) r"
unfolding SN_ars_def by simp

lemma SN_ars_E[elim]: "(SN::'a brel \<Rightarrow> bool) r \<Longrightarrow> ((!!(x::'a). SN r x) \<Longrightarrow> P) \<Longrightarrow> P"
unfolding SN_ars_def by simp

lemma SN_ars_E2[elim]:
 "(SN::'a brel \<Rightarrow> bool) r \<Longrightarrow> ((!!(x::'a) S. S 0 = x \<Longrightarrow> (\<exists>i. (S i,(S(Suc i))) \<notin> r)) \<Longrightarrow> P) \<Longrightarrow> P"
unfolding SN_ars_def SN_elt_def by auto

lemma WCR_elt_I[intro]:
 "(!!y z. (x,y) \<in> r \<Longrightarrow> (x,z) \<in> r \<Longrightarrow> (y,z) \<in> r^\<down>) \<Longrightarrow> (WCR::'a brel \<Rightarrow> 'a set) r x"
unfolding WCR_elt_def by simp

lemma WCR_elt_E[elim]:
 "(WCR::'a brel \<Rightarrow> 'a set) r x \<Longrightarrow> ((!!y z. (x,y) \<in> r \<Longrightarrow> (x,z) \<in> r \<Longrightarrow> (y,z) \<in> r^\<down>) \<Longrightarrow> P) \<Longrightarrow> P"
unfolding WCR_elt_def by simp

lemma WCR_ars_I[intro]: "(!!(x::'a). WCR r x) \<Longrightarrow> (WCR::'a brel \<Rightarrow> bool) r"
unfolding WCR_ars_def by simp

lemma WCR_ars_I2[intro]:
 "(!!(x::'a) y z. (x,y) \<in> r \<Longrightarrow> (x,z) \<in> r \<Longrightarrow> (y,z) \<in> r^\<down>) \<Longrightarrow> (WCR::'a brel \<Rightarrow> bool) r"
unfolding WCR_ars_def WCR_elt_def by auto

lemma WCR_ars_E[elim]: "(WCR::'a brel \<Rightarrow> bool) r \<Longrightarrow> ((!!(x::'a). WCR r x) \<Longrightarrow> P) \<Longrightarrow> P"
unfolding WCR_ars_def by simp

lemma WCR_ars_E2[elim]:
 "(WCR::'a brel \<Rightarrow> bool) r \<Longrightarrow> ((!!(x::'a) y z. (x,y) \<in> r \<Longrightarrow> (x,z) \<in> r \<Longrightarrow> (y,z) \<in> r^\<down>) \<Longrightarrow> P) \<Longrightarrow> P"
unfolding WCR_ars_def WCR_elt_def by auto

lemma WN_elt_I[intro]: "\<exists>y. (x,y) \<in> r^! \<Longrightarrow> (WN::'a brel \<Rightarrow> 'a set) r x"
unfolding WN_elt_def by simp

lemma WN_elt_E[elim]: "(WN::'a brel \<Rightarrow> 'a set) r x \<Longrightarrow> (\<exists>y.(x,y) \<in> r^! \<Longrightarrow> P) \<Longrightarrow> P"
unfolding WN_elt_def by simp

lemma WN_ars_I[intro]: "(!!(x::'a). WN r x) \<Longrightarrow> (WN::'a brel \<Rightarrow> bool) r"
unfolding WN_ars_def by simp

lemma WN_ars_E[elim]: "(WN::'a brel \<Rightarrow> bool) r \<Longrightarrow> ((!!(x::'a). WN r x) \<Longrightarrow> P) \<Longrightarrow> P"
unfolding WN_ars_def by simp

primrec S :: "'a brel \<Rightarrow> 'a \<Rightarrow> nat \<Rightarrow> 'a"
where "S r x 0 = x"
    | "S r x (Suc n) = (\<some> y. (S r x n, y) \<in> r)"

lemma SN_imp_WN: assumes "(SN::'a brel \<Rightarrow> bool) r" shows "(WN::'a brel \<Rightarrow> bool) r"
proof -
 have "\<forall>(x::'a). WN r x"
 proof
  fix x
  from assms have "\<not>(\<exists> S. S 0 = x \<and> (\<forall>i. (S i,(S(Suc i))) \<in> r))" by auto
  hence "\<exists>y. (x,y) \<in> {(x,y). (x,y) \<in> r^* \<and> y \<in> NF r}"
  proof (rule contrapos_np)
   let ?S = "\<lambda>i. S r x i"
   assume "\<not>(\<exists>y. (x,y) \<in> {(x,y). (x,y) \<in> r^* \<and> y \<in> NF r})"
   hence all:"\<forall>y. (x,y) \<in> r^* \<longrightarrow> (\<exists>z. (y,z) \<in> r)" by auto
   have "?S 0 = x \<and> (\<forall>i. (x,?S(Suc i)) \<in> r^Suc i \<and> (?S i,?S(Suc i)) \<in> r)"
   proof
    show "?S 0 = x" by simp
   next
    show "\<forall>i. (x,?S(Suc i)) \<in> r^Suc i \<and> (?S i,?S(Suc i)) \<in> r"
    proof
     fix i show "(x,?S(Suc i)) \<in> r^Suc i \<and> (?S i,?S(Suc i)) \<in> r"
     proof (induct i)
      case 0  thus ?case by (auto simp:all intro:someI_ex)
     next
      case (Suc n)
      hence IH: "(x,?S(Suc n)) \<in> r^Suc n" and "(?S n,?S(Suc n)) \<in> r" by auto
      hence "(x,?S(Suc n)) \<in> r^*" using rel_pow_imp_rtrancl by best
      with all have "\<exists>y. (?S(Suc n),y) \<in> r" by auto
      hence "(?S(Suc n),?S(Suc(Suc n))) \<in> r" by (auto intro:someI_ex)
      with IH show ?case using rel_pow_imp_rtrancl by auto
     qed
    qed
   qed
   thus "\<exists>S. S 0 = x \<and> (\<forall>i. (S i,S(Suc i)) \<in> r)" by best
  qed
  thus "WN r x" by auto
 qed
 thus ?thesis by auto
qed

hide const S

lemma UNC_imp_UNF:
 assumes "UNC r" shows "UNF r"
proof - {
 fix x y z assume "(x,y) \<in> r^!" and "(x,z) \<in> r^!"
 hence "(x,y) \<in> r^*" and "(x,z) \<in> r^*" and "y \<in> NF r" and "z \<in> NF r" by auto
 hence "(x,y) \<in> r^\<leftrightarrow>*" and "(x,z) \<in> r^\<leftrightarrow>*" by auto
 hence "(z,x) \<in> r^\<leftrightarrow>*" using convertible_sym unfolding sym_def by best
 with `(x,y) \<in> r^\<leftrightarrow>*` have "(z,y) \<in> r^\<leftrightarrow>*" using convertible_trans unfolding trans_def by best
 with `z \<in> NF r` and `y \<in> NF r` have "y = z" using assms by auto
} thus ?thesis by auto
qed

lemma joinable_NF_imp_eq:
 assumes "(x,y) \<in> r^\<down>" and "x \<in> NF r" and "y \<in> NF r"
 shows "x = y"
proof -
 from `(x,y) \<in> r^\<down>` obtain z where "(x,z)\<in>r^*" and "(z,y)\<in>(r\<inverse>)^*" unfolding joinable_def by auto
 hence "(y,z) \<in> r^*" unfolding rtrancl_converse by simp
 from `x \<in> NF r` have "(x,z) \<notin> r^+" using NF_no_trancl_step by best
 hence "x = z" using rtranclD[OF `(x,z) \<in> r^*`] by auto
 from `y \<in> NF r` have "(y,z) \<notin> r^+" using NF_no_trancl_step by best
 hence "y = z" using rtranclD[OF `(y,z) \<in> r^*`] by auto
 with `x = z` show ?thesis by simp
qed

lemma subsetI2:"(\<And>x y. (x,y) \<in> A \<Longrightarrow> (x,y) \<in> B) \<Longrightarrow> A \<subseteq> B" by auto

lemma CR_iff_meetable_subset_joinable:"(CR::'a brel \<Rightarrow> bool) r = (r^\<up> \<subseteq> r^\<down>)"
proof
 assume "CR r" show "r^\<up> \<subseteq> r^\<down>"
 proof (rule subsetI2)
  fix x y assume "(x,y) \<in> r^\<up>"
  then obtain z where "(z,x) \<in> r^*" and "(z,y) \<in> r^*" using meetable_D by best
  with `CR r` show "(x,y) \<in> r^\<down>" by auto
 qed
next
 assume "r^\<up> \<subseteq> r^\<down>" {
  fix x y z assume "(x,y) \<in> r^*" and "(x,z) \<in> r^*"
  hence "(y,z) \<in> r^\<up>" unfolding meetable_def rtrancl_converse by auto
  with `r^\<up> \<subseteq> r^\<down>` have "(y,z) \<in> r^\<down>" by auto
 } thus "CR r" by auto
qed

lemma CR_divergence_imp_joinable:
 assumes "(CR::'a brel \<Rightarrow> bool) r" and "(x,y) \<in> r^*" and "(x,z) \<in> r^*"
 shows "(y,z) \<in> r^\<down>"
proof -
 from `(x,y) \<in> r^*` and `(x,z) \<in> r^*` have "(y,z) \<in> r^\<up>" unfolding meetable_def rtrancl_converse by auto
 with assms show ?thesis unfolding CR_iff_meetable_subset_joinable by auto
qed

lemma joinable_imp_convertible: "r^\<down> \<subseteq> r^\<leftrightarrow>*"
proof (rule subsetI2)
 fix x z assume "(x,z) \<in> r^\<down>"
 then obtain y where "(x,y) \<in> r^*" and "(z,y) \<in> r^*" by auto
 hence "(x,y) \<in> r^\<leftrightarrow>*" and "(z,y) \<in> r^\<leftrightarrow>*" by auto
 from `(z,y) \<in> r^\<leftrightarrow>*` have "(y,z) \<in> r^\<leftrightarrow>*" using convertible_sym unfolding sym_def by best
 with `(x,y) \<in> r^\<leftrightarrow>*` show "(x,z) \<in> r^\<leftrightarrow>*" using convertible_trans unfolding trans_def by best
qed

lemma meetable_imp_convertible: "r^\<up> \<subseteq> r^\<leftrightarrow>*"
proof (rule subsetI2)
 fix y z assume "(y,z) \<in> r^\<up>"
 then obtain x where "(x,y) \<in> r^*" and "(x,z) \<in> r^*" by auto
 hence "(x,y) \<in> r^\<leftrightarrow>*" and "(x,z) \<in> r^\<leftrightarrow>*" by auto
 from `(x,y) \<in> r^\<leftrightarrow>*` have "(y,x) \<in> r^\<leftrightarrow>*" using convertible_sym unfolding sym_def by best
 with `(x,z) \<in> r^\<leftrightarrow>*` show "(y,z) \<in> r^\<leftrightarrow>*" using convertible_trans unfolding trans_def by best
qed

lemma CR_imp_UNF:
 assumes "(CR::'a brel \<Rightarrow> bool) r"
 shows "(UNF::'a brel \<Rightarrow> bool) r"
proof - {
fix x y z assume "(x,y) \<in> r^!" and "(x,z) \<in> r^!"
 hence "(x,y) \<in> r^*" and "y \<in> NF r" and "(x,z) \<in> r^*" and "z \<in> NF r" unfolding normalizable_def by auto
 from assms and `(x,y) \<in> r^*` and `(x,z) \<in> r^*` have "(y,z) \<in> r^\<down>" by (rule CR_divergence_imp_joinable)
 from this and `y \<in> NF r` and `z \<in> NF r` have "y = z" by (rule joinable_NF_imp_eq)
} thus ?thesis by auto
qed

lemma CR_iff_convertible_imp_joinable:"(CR::'a brel \<Rightarrow> bool) r = (r^\<leftrightarrow>* \<subseteq> r^\<down>)"
proof (intro iffI subsetI2)
 fix x y assume "CR r" and "(x,y) \<in> r^\<leftrightarrow>*"
 then obtain n where "(x,y) \<in> (r^\<leftrightarrow>)^n" unfolding convertible_def rtrancl_is_UN_rel_pow by auto
 thus "(x,y) \<in> r^\<down>"
 proof (induct n arbitrary: x)
  case 0
  assume "(x,y) \<in> r^\<leftrightarrow> ^ 0" hence "x = y" by simp show ?case unfolding `x = y` by (rule joinability_refl')
 next
  case (Suc n)
  from `(x,y) \<in> r^\<leftrightarrow> ^ Suc n` obtain  z where "(x,z) \<in> r^\<leftrightarrow>" and "(z,y) \<in> r^\<leftrightarrow> ^ n" using rel_pow_Suc_D2 by best
  with Suc have "(z,y) \<in> r^\<down>" by simp
  from `(x,z) \<in> r^\<leftrightarrow>`
  show ?case
  proof (rule symmc_E)
   assume "(x,z) \<in> r" with `(z,y) \<in> r^\<down>` show ?thesis by auto
  next
   assume "(z,x) \<in> r"
   hence "(z,x) \<in> r^*" by simp
   from `(z,y) \<in> r^\<down>` obtain z' where "(z,z') \<in> r^*" and "(y,z') \<in> r^*" by auto
   from `CR r` and `(z,x) \<in> r^*` and `(z,z') \<in> r^*` have "(x,z') \<in> r^\<down>" by (rule CR_divergence_imp_joinable)
   then obtain x' where "(x,x') \<in> r^*" and "(z',x') \<in> r^*" by auto
   with `(y,z') \<in> r^*` show ?thesis by auto
  qed
 qed
next
 assume "r^\<leftrightarrow>* \<subseteq> r^\<down>" thus "CR r" unfolding CR_iff_meetable_subset_joinable using meetable_imp_convertible by auto
qed

lemma CR_imp_convertible_iff_joinable:
 assumes "(CR::'a brel \<Rightarrow> bool) r"
 shows "r^\<leftrightarrow>* = r^\<down>"
proof
 show "r^\<leftrightarrow>* \<subseteq> r^\<down>" using CR_iff_convertible_imp_joinable assms by auto
next
 show "r^\<down> \<subseteq> r^\<leftrightarrow>*" by (rule joinable_imp_convertible)
qed

lemma joinable_sym: "sym(r^\<down>)" unfolding sym_def by auto

lemma CR_joinable_left_I[intro]:
 assumes "(CR::'a brel \<Rightarrow> bool) r" and "(x,y) \<in> r^*" and "(x,z) \<in> r^\<down>"
 shows "(y,z) \<in> r^\<down>"
proof -
 from `(x,z) \<in> r^\<down>` obtain x' where "(x,x') \<in> r^*" and "(z,x') \<in> r^\<down>" by auto
 from `CR r` and `(x,x') \<in> r^*` and `(x,y) \<in> r^*` have "(x,y) \<in> r^\<down>" by auto
 hence "(y,x) \<in> r^\<down>" using joinable_sym unfolding sym_def by best
 from `CR r` have "r^\<leftrightarrow>* = r^\<down>" by (rule CR_imp_convertible_iff_joinable)
 from `(y,x) \<in> r^\<down>` and `(x,z) \<in> r^\<down>` show ?thesis using convertible_trans unfolding trans_def `r^\<leftrightarrow>* = r^\<down>`[symmetric] by best
qed

lemma CR_joinable_right_I[intro]:
 assumes "(CR::'a brel \<Rightarrow> bool) r" and "(x,y) \<in> r^\<down>" and "(y,z) \<in> r^*"
 shows "(x,z) \<in> r^\<down>"
proof -
 have "r^\<leftrightarrow>* = r^\<down>" by (rule CR_imp_convertible_iff_joinable[OF `CR r`])
 from `(y,z) \<in> r^*` have "(y,z) \<in> r^\<leftrightarrow>*" by auto
 with `(x,y) \<in> r^\<down>` show ?thesis unfolding `r^\<leftrightarrow>* = r^\<down>`[symmetric] using convertible_trans unfolding trans_def by fast
qed

lemma NF_not_suc:
 assumes "(x,y) \<in> r^*" and "x \<in> NF r"
 shows "x = y"
proof -
 from `x \<in> NF r` have "\<forall>y. (x,y) \<notin> r" using NF_no_step by auto
 hence "x \<notin> Domain r" unfolding Domain_def by simp
 from `(x,y) \<in> r^*` show ?thesis unfolding Not_Domain_rtrancl[OF `x \<notin> Domain r`] by simp
qed

lemma semi_complete_imp_convertible_iff_same_NF:
 assumes "semi_complete r"
 shows "((x,y) \<in> r^\<leftrightarrow>*) = (\<forall>u v. (x,u) \<in> r^! \<and> (y,v) \<in> r^! \<longrightarrow> u = v)"
proof -
 from assms have "WN r" and "CR r" by auto
 hence "r^\<leftrightarrow>* = r^\<down>" using CR_imp_convertible_iff_joinable by auto
 show ?thesis
 proof
  assume "(x,y) \<in> r^\<leftrightarrow>*"
  from `(x,y) \<in> r^\<leftrightarrow>*` have "(x,y) \<in> r^\<down>" unfolding `r^\<leftrightarrow>* = r^\<down>` .
  show "\<forall>u v. (x,u) \<in> r^! \<and> (y,v) \<in> r^! \<longrightarrow> u = v"
  proof (intro allI impI, elim conjE)
   fix u v assume "(x,u) \<in> r^!" and "(y,v) \<in> r^!"
   hence "(x,u) \<in> r^*" and "(y,v) \<in> r^*" and "u \<in> NF r" and "v \<in> NF r" by auto
   from `CR r` and `(x,u) \<in> r^*` and `(x,y) \<in> r^\<down>` have "(u,y) \<in> r^\<down>" by auto
   hence "(y,u) \<in> r^\<down>" using joinable_sym unfolding sym_def by best
   with `(x,y) \<in> r^\<down>` have "(x,u) \<in> r^\<down>" unfolding `r^\<leftrightarrow>* = r^\<down>`[symmetric] using convertible_trans unfolding trans_def by best
   from `CR r` and `(x,y) \<in> r^\<down>` and `(y,v) \<in> r^*` have "(x,v) \<in> r^\<down>" by auto
   hence "(v,x) \<in> r^\<down>" using joinable_sym unfolding sym_def by best
   with `(x,u) \<in> r^\<down>` have "(v,u) \<in> r^\<down>" unfolding `r^\<leftrightarrow>* = r^\<down>`[symmetric] using convertible_trans unfolding trans_def by best
   then obtain v' where "(v,v') \<in> r^*" and "(u,v') \<in> r^*" by auto
   from `(u,v') \<in> r^*` and `u \<in> NF r` have "u = v'" by (rule NF_not_suc)
   from `(v,v') \<in> r^*` and `v \<in> NF r` have "v = v'" by (rule NF_not_suc)
   thus "u = v" unfolding `u = v'` by simp
  qed
 next
  assume equal_NF:"\<forall>u v. (x,u) \<in> r^! \<and> (y,v) \<in> r^! \<longrightarrow> u = v"
  from `WN r` obtain u where "(x,u) \<in> r^!" by best
  from `WN r` obtain v where "(y,v) \<in> r^!" by best
  from `(x,u) \<in> r^!` and `(y,v) \<in> r^!` have "u = v" using equal_NF by simp
  from `(x,u) \<in> r^!` and `(y,v) \<in> r^!` have "(x,v) \<in> r^*" and "(y,v) \<in> r^*" unfolding `u = v` by auto
  hence "(x,v) \<in> r^\<leftrightarrow>*" and "(y,v) \<in> r^\<leftrightarrow>*" by auto
  from `(y,v) \<in> r^\<leftrightarrow>*` have "(v,y) \<in> r^\<leftrightarrow>*" using convertible_sym unfolding sym_def by best
  with `(x,v) \<in> r^\<leftrightarrow>*` show "(x,y) \<in> r^\<leftrightarrow>*" using convertible_trans unfolding trans_def by best
 qed
qed

lemma CR_imp_UNC:
 assumes "CR r" shows "UNC r"
proof - {
 fix x y assume "x \<in> NF r" and "y \<in> NF r" and "(x,y) \<in> r^\<leftrightarrow>*"
 have "r^\<leftrightarrow>* = r^\<down>" by (rule CR_imp_convertible_iff_joinable[OF `CR r`])
 from `(x,y) \<in> r^\<leftrightarrow>*` have "(x,y) \<in> r^\<down>" unfolding `r^\<leftrightarrow>* = r^\<down>` by simp
 then obtain x' where "(x,x') \<in> r^*" and "(y,x') \<in> r^*" by best
 from `(x,x') \<in> r^*` and `x \<in> NF r` have "x = x'" by (rule NF_not_suc)
 from `(y,x') \<in> r^*` and `y \<in> NF r` have "y = x'" by (rule NF_not_suc)
 hence "x = y" unfolding `x = x'` by simp
} thus ?thesis by auto
qed

lemma WN_UNF_imp_CR:
 assumes "WN r" and "UNF r" shows "(CR::'a brel \<Rightarrow> bool) r"
proof - {
 fix x y z assume "(x,y) \<in> r^*" and "(x,z) \<in> r^*"
 from `WN r` obtain y' where "(y,y') \<in> r^!" unfolding WN_ars_def WN_elt_def by best
 with `(x,y) \<in> r^*` have "(x,y') \<in> r^!" by auto
 from `WN r` obtain z' where "(z,z') \<in> r^!" unfolding WN_ars_def WN_elt_def by best
 with `(x,z) \<in> r^*` have "(x,z') \<in> r^!" by auto
 with `(x,y') \<in> r^!` have "y' = z'" using `UNF r` unfolding UNF_ars_def UNF_elt_def by auto
 from `(y,y') \<in> r^!` and `(z,z') \<in> r^!` have "(y,z) \<in> r^\<down>" unfolding `y' = z'` by auto
} thus ?thesis by auto
qed

definition diamond :: "'a brel \<Rightarrow> bool" ("\<diamond>") where
 "\<diamond> r \<equiv> (r O r\<inverse>) \<subseteq> (r\<inverse> O r)"

lemma diamond_I[intro]: "(r O r\<inverse>) \<subseteq> (r\<inverse> O r) \<Longrightarrow> \<diamond> r" unfolding diamond_def by simp

lemma diamond_E[elim]: "\<diamond> r \<Longrightarrow> ((r O r\<inverse>) \<subseteq> (r\<inverse> O r) \<Longrightarrow> P) \<Longrightarrow> P" unfolding diamond_def by simp

lemma diamond_imp_semi_confluence:
 assumes "\<diamond> r" shows "(r^* O r\<inverse>) \<subseteq> r^\<down>"
proof (rule subsetI2)
 fix y z assume "(y,z) \<in> r^* O r\<inverse>"
 then obtain x where "(x,y) \<in> r" and "(x,z) \<in> r^*" by best
 then obtain n where "(x,z) \<in> r^n" using rtrancl_imp_UN_rel_pow by best
 with `(x,y) \<in> r` show "(y,z) \<in> r^\<down>"
 proof (induct n arbitrary: x z y)
  case 0 thus ?case by auto
 next
  case (Suc n)
  from `(x,z) \<in> r^Suc n` obtain x' where "(x,x') \<in> r" and "(x',z) \<in> r^n" using rel_pow_Suc_D2 by best
  with `(x,y) \<in> r` have "(y,x') \<in> (r O r\<inverse>)" by auto
  with `\<diamond> r` have "(y,x') \<in> (r\<inverse> O r)" by auto
  then obtain y' where "(x',y') \<in> r" and "(y,y') \<in> r" by best
  with Suc and `(x',z) \<in> r^n` have "(y',z) \<in> r^\<down>" by auto
  with `(y,y') \<in> r` show ?case by auto
 qed
qed

lemma semi_confluence_imp_CR:
 assumes "(r^* O r\<inverse>) \<subseteq> r^\<down>"
 shows "(CR::'a brel \<Rightarrow> bool) r"
proof - {
 fix x y z assume "(x,y) \<in> r^*" and "(x,z) \<in> r^*"
 then obtain n where "(x,z) \<in> r^n" using rtrancl_imp_UN_rel_pow by best
 with `(x,y) \<in> r^*` have "(y,z) \<in> r^\<down>"
 proof (induct n arbitrary: x y z)
  case 0 thus ?case by auto
 next
  case (Suc n)
  from `(x,z) \<in> r^Suc n` obtain x' where "(x,x') \<in> r" and "(x',z) \<in> r^n" using rel_pow_Suc_D2 by best
  from `(x,x') \<in> r` and `(x,y) \<in> r^*` have "(x',y) \<in> (r^* O r\<inverse>)" by auto
  with assms have "(x',y) \<in> r^\<down>" by auto
  then obtain y' where "(x',y') \<in> r^*" and "(y,y') \<in> r^*" by best
  with Suc and `(x',z) \<in> r^n` have "(y',z) \<in> r^\<down>" by simp
  then obtain u where "(z,u) \<in> r^*" and "(y',u) \<in> r^*" by best
  from `(y,y') \<in> r^*` and `(y',u) \<in> r^*` have "(y,u) \<in> r^*" by auto
  with `(z,u) \<in> r^*` show ?case by best
 qed
} thus ?thesis by auto
qed
 
lemma diamond_imp_CR: assumes "\<diamond> r" shows "(CR::'a brel \<Rightarrow> bool) r"
using assms by (rule diamond_imp_semi_confluence[THEN semi_confluence_imp_CR])

lemma diamond_imp_CR':
 assumes "\<diamond> s" and "r \<subseteq> s" and "s \<subseteq> r^*" shows "(CR::'a brel \<Rightarrow> bool) r" unfolding CR_iff_meetable_subset_joinable
proof -
 from `\<diamond> s` have "CR s" by (rule diamond_imp_CR)
 hence "s^\<up> \<subseteq> s^\<down>" unfolding CR_iff_meetable_subset_joinable by simp
 from `r \<subseteq> s` have "r^* \<subseteq> s^*" by (rule rtrancl_mono)
 from `s \<subseteq> r^*` have "s^* \<subseteq> (r^*)^*" by (rule rtrancl_mono)
 hence "s^* \<subseteq> r^*" by simp
 with `r^* \<subseteq> s^*` have "r^* = s^*" by simp
 show "r^\<up> \<subseteq> r^\<down>" unfolding meetable_def joinable_def rtrancl_converse `r^* = s^*`
  unfolding rtrancl_converse[symmetric] meetable_def[symmetric] joinable_def[symmetric] by (rule `s^\<up> \<subseteq> s^\<down>`)
qed

subsubsection {* Exercises *}
text {* ALN, Exercise~1.9.a *}
lemma NF_imp_complete: 
 assumes "x \<in> NF r" shows "complete r x"
proof - {
  from assms have "\<forall>y. (x,y) \<notin> r" by (rule NF_no_step)
  hence "\<not>(\<exists>S. S 0 = x \<and> (\<forall>i. (S i,S(Suc i)) \<in> r))" by auto
 }
 hence "SN r x" by auto {
  fix y z assume "(x,y) \<in> r^* \<and> (x,z) \<in> r^*"
  with `x \<in> NF r` have "x = y" and "x = z" by (auto simp:NF_not_suc)
  hence "(y,z) \<in> r^\<down>" by auto
 }
 hence "CR r x" by auto
 with `SN r x` show ?thesis by auto
qed

text {* ALN, Exercise~1.9.b *}
lemma semi_complete_imp_UNF_exists:
 assumes "semi_complete r"
 shows "\<exists>y. (x,y) \<in> r^! \<and> UNF r x"
proof -
 from assms have "WN r x" and "CR r x" by auto
 {
  fix y z assume "(x,y) \<in> r^!" and "(x,z) \<in> r^!"
  hence "(x,y) \<in> r^*" and "(x,z) \<in> r^*" and "y \<in> NF r" and "z \<in> NF r" by auto
  with `CR r x` have "(y,z) \<in> r^\<down>" by auto
  with `y \<in> NF r` and `z \<in> NF r` have "y = z" by (auto intro: joinable_NF_imp_eq)
 } hence "UNF r x" by auto
 from `WN r x` have "\<exists>y. (x,y) \<in> r^!" by auto
 with `UNF r x` show ?thesis by simp
qed

lemma UNF_exists_imp_convertible_iff_joinable_to_NF:
 assumes "\<forall>x. \<exists>y. (x,y) \<in> r^! \<and> UNF r x"
 shows "r^\<leftrightarrow>* = ((r^!)\<inverse> O r^!)"
proof
 show "r^\<leftrightarrow>* \<subseteq> (r^!)\<inverse> O r^!"
 proof (rule subsetI2)
  fix x y assume "(x,y) \<in> r^\<leftrightarrow>*"
  hence "(x,y) \<in> (r^\<leftrightarrow>)^*" by auto
  then obtain n where "(x,y) \<in> (r^\<leftrightarrow>)^n" using rtrancl_imp_UN_rel_pow by best
  thus "(x,y) \<in> (r^!)\<inverse> O r^!"
  proof (induct n arbitrary: x y)
   case 0 hence "x = y" by simp
   from assms obtain x' where "(x,x') \<in> r^!" and "(x',x) \<in> (r^!)\<inverse>" by best
   thus ?case unfolding `x = y` by auto
  next
   case (Suc n)
   from `(x,y) \<in> (r^\<leftrightarrow>)^Suc n`
    obtain x' where "(x,x') \<in> r^\<leftrightarrow>" and "(x',y) \<in> (r^\<leftrightarrow>)^n" using rel_pow_Suc_D2 by best
   with Suc have "(x',y) \<in> (r^!)\<inverse> O r^!" by simp
   then obtain u where "(x',u) \<in> r^!" and "(y,u) \<in> r^!" by auto
   from assms have "UNF r x" and "UNF r x'" by auto
   from assms obtain v where "(x,v) \<in> r^!" by best
   from `(x,x') \<in> r^\<leftrightarrow>` show ?case
   proof (rule symmc_E)
    assume "(x,x') \<in> r"
    with `(x',u) \<in> r^!` have "(x,u) \<in> r^!" by auto
    with `UNF r x` and `(x,v) \<in> r^!` have "v = u" by auto
    from `(x,v) \<in> r^!` and `(y,u) \<in> r^!` show ?thesis unfolding `v = u` by auto
   next
    assume "(x',x) \<in> r"
    with `(x,v) \<in> r^!` have "(x',v) \<in> r^!" by auto
    with `UNF r x'` and `(x',u) \<in> r^!` have "u = v" by auto
    from `(x,v) \<in> r^!` and `(y,u) \<in> r^!` show ?thesis unfolding `u = v` by auto
   qed
  qed
 qed
next
 show "(r^!)\<inverse> O r^! \<subseteq> r^\<leftrightarrow>*"
 proof (rule subsetI2)
  fix x y assume "(x,y) \<in> (r^!)\<inverse> O r^!"
  then obtain x' where "(x,x') \<in> r^!" and "(y,x') \<in> r^!" by auto
  hence "(x,x') \<in> r^*" and "(y,x') \<in> r^*" by auto
  hence "(x,x') \<in> r^\<leftrightarrow>*" and "(y,x') \<in> r^\<leftrightarrow>*" by auto
  hence "(x',y) \<in> r^\<leftrightarrow>*" using convertible_sym unfolding sym_def by fast
  with `(x,x') \<in> r^\<leftrightarrow>*` show "(x,y) \<in> r^\<leftrightarrow>*" using convertible_trans unfolding trans_def by fast
 qed
qed

lemma normalizable_trancl_absorb: "(r^!)^+ = r^!"
proof (rule equalityI)
 show "(r^!)^+ \<subseteq> r^!"
 proof (rule subsetI2)
  fix x y assume "(x,y) \<in> (r^!)^+" thus "(x,y) \<in> r^!" by induct auto
 qed
next
 show "r^! \<subseteq> (r^!)^+"
 proof (rule subsetI2)
  fix x y assume "(x,y) \<in> r^!" thus "(x,y) \<in> (r^!)^+" by auto
 qed
qed

lemma convertible_iff_joinable_to_NF_imp_semi_complete:
 assumes "r^\<leftrightarrow>* = ((r^!)\<inverse> O r^!)"
 shows "semi_complete r"
proof - {
 fix x 
 {
  have "(x,x) \<in> r^\<leftrightarrow>*" by auto
  hence "\<exists>y. (x,y) \<in> r^!" unfolding assms by auto
 }
 hence "(WN::'a brel \<Rightarrow> 'a set) r x" by auto
 {
  fix y z assume "(x,y) \<in> r^*" and "(x,z) \<in> r^*"
  hence "(x,y) \<in> r^\<leftrightarrow>*" and "(x,z) \<in> r^\<leftrightarrow>*" by auto
  from `(x,y) \<in> r^\<leftrightarrow>*` have "(y,x) \<in> r^\<leftrightarrow>*" using convertible_sym unfolding sym_def by fast
  with `(x,z) \<in> r^\<leftrightarrow>*` have "(y,z) \<in> r^\<leftrightarrow>*" using convertible_trans unfolding trans_def by fast
  then obtain y' where "(y,y') \<in> r^!" and "(z,y') \<in> r^!" unfolding assms by best
  hence "(y,y') \<in> r^*" and "(z,y') \<in> r^*" by auto
  hence "(y,z) \<in> r^\<down>" by auto
 }
 hence "(CR::'a brel \<Rightarrow> 'a set) r x" by auto
 with `WN r x` have "semi_complete r x" by auto
} thus ?thesis by auto
qed

text {* ALN, Exercise~1.9.c *}
lemma WN_iff_NF_of_normalizable_empty: "WN r = (NF(r^!) = {})"
proof
 assume "WN r"
 {
  fix x assume "x \<in> NF(r^!)"
  hence "\<forall>y. (x,y) \<notin> r^!" by auto
  from `WN r` have "\<exists>y. (x,y) \<in> r^!" by auto
  with `\<forall>y. (x,y) \<notin> r^!` have "x \<in> {}" by simp
 }
 hence "NF(r^!) \<subseteq> {}" by auto
 thus "NF(r^!) = {}" unfolding subset_empty by simp
next
 assume "NF(r^!) = {}"
 {
  fix x have "\<exists>y. (x,y) \<in> r^!"
  proof (rule ccontr)
   assume "\<not>(\<exists>y. (x,y) \<in> r^!)"
   hence "x \<in> NF(r^!)" by auto
   with `NF(r^!) = {}` show False by simp
  qed
  hence "WN r x" by auto
 }
 thus "(WN::'a brel \<Rightarrow> bool) r" by auto
qed

text {* ALN, Exercise~1.10.a *}
lemma SN_imp_SN_reduct:
 assumes "(SN::'a brel \<Rightarrow> 'a set) r x" and "(x,y) \<in> r^*"
 shows "SN r y"
proof - {
 assume a:"\<exists>S. S 0 = y \<and> (\<forall>i. (S i,S(Suc i)) \<in> r)"
 from `(x,y) \<in> r^*` obtain n where "(x,y) \<in> r^n" using rtrancl_imp_UN_rel_pow by best
 hence "\<not> SN r x"
 proof (induct n arbitrary: x)
  case 0 hence "x = y" by simp
  show ?case using a unfolding `x = y` by auto
 next
  case (Suc n)
  from `(x,y) \<in> r^Suc n` obtain x' where "(x,x') \<in> r" and "(x',y) \<in> r^n" using rel_pow_Suc_D2 by best
  with Suc have "\<not> SN r x'" by simp
  then obtain S where "S 0 = x'" and "\<forall>i. (S i,S(Suc i)) \<in> r" by auto
  let ?S = "\<lambda>i. case i of 0 \<Rightarrow> x | Suc i \<Rightarrow> S i"
  have "?S 0 = x" by simp
  have "\<forall>i. (?S i,?S(Suc i)) \<in> r"
  proof
   fix i show "(?S i,?S(Suc i)) \<in> r" by (induct i) (auto simp: `S 0 = x'` `(x,x') \<in> r` `\<forall>i. (S i,S(Suc i)) \<in> r`)
  qed
  with `?S 0 = x` have "\<exists>S. S 0 = x \<and> (\<forall>i. (S i,S(Suc i)) \<in> r)" by best
  thus ?case by auto
 qed
} with `SN r x` show ?thesis by auto
qed

text {* ALN, Exercise~1.10.b *}
lemma CR_imp_CR_reduct:
 assumes "(CR::'a brel \<Rightarrow> 'a set) r x" and "(x,y) \<in> r^*"
 shows "CR r y"
proof - {
 fix y' z' assume "(y,y') \<in> r^*" and "(y,z') \<in> r^*"
 with `(x,y) \<in> r^*` have "(x,y') \<in> r^*" and "(x,z') \<in> r^*" by auto
 with `CR r x` have "(y',z') \<in> r^\<down>" by auto
} thus ?thesis by auto
qed

lemma UNF_imp_UNF_reduct:
 assumes "(UNF::'a brel \<Rightarrow> 'a set) r x" and "(x,y) \<in> r^*"
 shows "UNF r y"
proof - {
 fix y' z' assume "(y,y') \<in> r^!" and "(y,z') \<in> r^!"
 with `(x,y) \<in> r^*` have "(x,y') \<in> r^!" and "(x,z') \<in> r^!" by auto
 with `UNF r x` have "y' = z'" by auto
} thus ?thesis by auto
qed

lemma complete_imp_complete_reduct:
 assumes "(complete::'a brel \<Rightarrow> 'a set) r x" and "(x,y) \<in> r^*"
 shows "complete r y"
proof - {
 from `complete r x` have "SN r x" and "CR r x" by auto
 have "SN r y" and "CR r y"
  using SN_imp_SN_reduct[OF `SN r x` `(x,y) \<in> r^*`] CR_imp_CR_reduct[OF `CR r x` `(x,y) \<in> r^*`] by auto
} thus ?thesis by auto
qed

text {* ALN, Exercise~1.10.c *}
lemma WN_imp_WN_anc:
 assumes "(WN::'a brel \<Rightarrow> 'a set) r x" and "(y,x) \<in> r^*"
 shows "WN r y"
proof - 
 from `WN r x` obtain z where"(x,z) \<in> r^!"  by auto
 with `(y,x) \<in> r^*` show ?thesis by auto
qed

text {* ALN, Exercise~1.13 *}
definition NFP :: "'a brel \<Rightarrow> bool" where
 "NFP r \<equiv> (\<forall>(x::'a). WN r x \<longrightarrow> CR r x)"

lemma NFP_ars_I[intro]: "(!!(x::'a). WN r x \<Longrightarrow> CR r x) \<Longrightarrow> (NFP::'a brel \<Rightarrow> bool) r" 
unfolding NFP_def by simp

lemma NFP_ars_E[elim]: "(NFP::'a brel \<Rightarrow> bool) r \<Longrightarrow> ((!!(x::'a). WN r x \<Longrightarrow> CR r x) \<Longrightarrow> P) \<Longrightarrow> P"
unfolding NFP_def by simp


text {* ALN, Exercise~1.13 a) *}
lemma CR_imp_NFP:
 assumes "CR (r::'a brel)"
 shows "NFP r"
proof
 fix x :: 'a
 assume "WN r x"
 show "CR r x :: bool" using assms unfolding CR_ars_def by auto
qed

text {* ALN, Exercise~1.13 c) *}
text {* If A has the NFP then \leftarrow\cdot \to^! \subseteq \to^!*}
lemma one_to_two:
 assumes "NFP (r::'a brel)"
 shows "(r^! O r\<inverse>) \<subseteq> r^!"
proof (rule subsetI2)
 fix y z assume "(y, z) \<in> (r^! O r\<inverse>)"
 then obtain x where "(x,y) \<in> r" and "(x,z) \<in> r^!" by best
 with assms have "CR r x" and "WN r x" unfolding NFP_def by auto
 from `(x,y) \<in> r` and `(x,z) \<in> r^!` have " (x,y) \<in> r^*" and "(x,z) \<in> r^*" by auto
 with `CR r x` obtain u where "(y,u) \<in> r^*" and "(z,u) \<in> r^*" by best
 from `(x,z) \<in> r^!` have "z \<in> NF r" by auto
 with `(z,u) \<in> r^*` have "z=u" by (rule NF_not_suc)
 with `z \<in> NF r` and `(y,u) \<in> r^*` show  "(y,z) \<in> r^!" by auto  
qed

text {* If \leftarrow\cdot \to^! \subseteq \to^! then any element
convertible to a NF rewrites to that NF *}
lemma two_to_three:
 assumes "((r::'a brel)^! O r\<inverse>) \<subseteq> r^!"
 shows "\<forall>x y. ((x,y) \<in> r^\<leftrightarrow>* \<and> y \<in> NF r \<longrightarrow> (x,y) \<in> r^* )"
proof (intro allI)
 fix x y 
 show "(x, y) \<in> r^\<leftrightarrow>* \<and> y \<in> NF r \<longrightarrow> (x, y) \<in> r^*"
 proof
  assume "(x, y) \<in> r^\<leftrightarrow>* \<and> y \<in> NF r"
  hence "(x,y) \<in> r^\<leftrightarrow>*" and "y  \<in> NF r" by auto
  then obtain n where "(x,y) \<in> (r^\<leftrightarrow>)^n" unfolding convertible_def rtrancl_is_UN_rel_pow by auto
  thus "(x, y) \<in> r^*"
  proof (induct n arbitrary: x)
   case 0
   assume "(x,y) \<in> r^\<leftrightarrow> ^ 0" hence "x = y" by simp show ?case unfolding `x = y` by simp
  next
   case (Suc n)
   from `(x,y) \<in> r^\<leftrightarrow> ^ Suc n` obtain  z where "(x,z) \<in> r^\<leftrightarrow>" and "(z,y) \<in> r^\<leftrightarrow> ^ n" using rel_pow_Suc_D2 by best
   with Suc have "(z,y) \<in> r^*" by simp
   with `y \<in> NF r` have  "(z,y) \<in> r^!" by auto
   from `(x,z) \<in> r^\<leftrightarrow>` show ?case
   proof (rule symmc_E) 
    assume "(x,z) \<in> r" with `(z,y) \<in> r^!` show ?thesis by auto
   next
    assume "(z,x) \<in> r"
    with `(z,y) \<in> r^!` have "(x,y) \<in> (r^! O r\<inverse>)" by best
    with assms show ?thesis by auto
   qed
  qed
 qed
qed

text {* If any element convertible to a NF rewrites to that NF 
then r has the NFP *}
lemma three_to_one:
 assumes "\<forall>x y. ((x,y) \<in> (r::'a brel)^\<leftrightarrow>* \<and> y \<in> NF r \<longrightarrow> (x,y) \<in> r^* )"  
 shows "NFP r"
proof (rule NFP_ars_I)
 fix x::'a
 assume "WN r x"
 then obtain z where "(x, z) \<in> r^!" unfolding WN_elt_def by auto
 hence "(x, z)  \<in> r^\<leftrightarrow>*" and "z \<in> NF r" by auto
 show "CR r x" 
 proof (rule CR_elt_I)
 fix u v
 assume  "(x,u) \<in> r^*"  and "(x,v) \<in> r^*"
 hence "(u, x) \<in> r^\<leftrightarrow>*" using convertible_sym unfolding sym_def by best
 with `(x, z)  \<in> r^\<leftrightarrow>*` have "(u, z) \<in> r^\<leftrightarrow>*" using convertible_trans unfolding trans_def by best
 from assms and `(u, z) \<in> r^\<leftrightarrow>*` and `z \<in> NF r` have "(u,z) \<in> r^*" by auto
 from `(x, v) \<in> r^*` have "(v, x) \<in> r^\<leftrightarrow>*" using convertible_sym unfolding sym_def by best 
 with `(x, z)  \<in> r^\<leftrightarrow>*` have "(v, z) \<in> r^\<leftrightarrow>*" using convertible_trans unfolding trans_def by best
 from assms and `(v, z) \<in> r^\<leftrightarrow>*` and `z \<in> NF r` have "(v,z) \<in> r^*" by auto
 from `(u, z) \<in> r^*` `(v,z) \<in> r^*` show "(u,v) \<in> r^\<down>" by (rule joinable_I)
 qed
qed

text {* ALN, Exercise~1.13 d) *}

text {* NFP implies UNC and thus also UNF. The reverse implications do not hold. *}
lemma NFP_imp_UNC:
 assumes "NFP r"
 shows "UNC r"
proof
 fix x y
 assume "x \<in> NF r" and "y \<in> NF r" and "(x, y) \<in> r^\<leftrightarrow>*"
 show "x=y"
 proof -
  from `NFP r` have  "((r::'a brel)^! O r\<inverse>) \<subseteq> r^!" by (rule one_to_two)
  with `(x, y) \<in> r^\<leftrightarrow>*` and `y \<in> NF r` have "(x,y) \<in> r^*" using two_to_three by auto
  with this and `x \<in> NF r` show "x=y" using NF_not_suc by auto
 qed
qed

lemma NFP_imp_UN:
 assumes "NFP r"
 shows "UNF r"
proof -
 from assms have "UNC r" by (rule NFP_imp_UNC)
 thus ?thesis by (rule UNC_imp_UNF)
qed



subsection {* Newman's Lemma *}
lemma wf_imp_SN:
 assumes "wf(r\<inverse>)" shows "SN r"
proof - {
 fix x
 let ?P = "\<lambda>x. \<not>(\<exists>S. S 0 = x \<and> (\<forall>i. (S i,S(Suc i)) \<in> r))"
 from `wf(r\<inverse>)` have "?P x"
 proof induct
  case (less x)
  hence IH: "\<And>y. (x,y) \<in> r \<Longrightarrow> ?P y" by auto
  show "?P x"
  proof (rule ccontr)
   assume "\<not> ?P x"
   then obtain S where "S 0 = x" and "\<forall>i. (S i,S(Suc i)) \<in> r" by auto
   hence "(S 0,S 1) \<in> r" by auto
   with IH have "?P(S 1)" unfolding `S 0 = x` by auto
   with `\<forall>i. (S i,S(Suc i)) \<in> r` show False by auto
  qed
 qed
 hence "SN r x" by auto
} thus ?thesis by auto
qed

primrec T :: "('a \<Rightarrow> 'a) \<Rightarrow> 'a \<Rightarrow> nat \<Rightarrow> 'a"
where "T f x 0 = x"
    | "T f x (Suc n) = f (T f x n)" 

lemma SN_imp_minimal:
 assumes "(SN::'a brel \<Rightarrow> bool) r"
 shows "\<forall>Q x. x \<in> Q \<longrightarrow> (\<exists>z\<in>Q. \<forall>y. (z,y) \<in> r \<longrightarrow> y \<notin> Q)"
proof (rule ccontr)
 assume "\<not>(\<forall>Q x. x \<in> Q \<longrightarrow> (\<exists>z\<in>Q. \<forall>y. (z,y) \<in> r \<longrightarrow> y \<notin> Q))"
 then obtain Q x where "x \<in> Q" and "\<forall>z\<in>Q. \<exists>y. (z,y) \<in> r \<and> y \<in> Q" by auto
 hence "\<forall>z. \<exists>y. z \<in> Q \<longrightarrow> (z,y) \<in> r \<and> y \<in> Q" by auto
 hence "\<exists>f. \<forall>x. x \<in> Q \<longrightarrow> (x,f x) \<in> r \<and> f x \<in> Q" by (rule choice)
 then obtain f where a:"\<forall>x. x \<in> Q \<longrightarrow> (x,f x) \<in> r \<and> f x \<in> Q" (is "\<forall>x. ?P x") by best
 let ?S = "\<lambda>i. T f x i"
 have "?S 0 = x" by simp
 have "\<forall>i. (?S i,?S(Suc i)) \<in> r \<and> ?S(Suc i) \<in> Q"
 proof
  fix i show "(?S i,?S(Suc i)) \<in> r \<and> ?S(Suc i) \<in> Q" by (induct i) (auto simp:`x \<in> Q` a)
 qed
 with `?S 0 = x` have "\<exists>S. S 0 = x \<and> (\<forall>i. (S i,S(Suc i)) \<in> r)" by fast
 with assms show False by auto
qed

lemma SN_elt_imp_elt_minimal:
 assumes "SN r x"
 shows "\<forall>Q. x \<in> Q \<longrightarrow> (\<exists>z\<in>Q. \<forall>y. (z,y) \<in> r \<longrightarrow> y \<notin> Q)"
proof (rule ccontr)
 assume "\<not>(\<forall>Q. x \<in> Q \<longrightarrow> (\<exists>z\<in>Q. \<forall>y. (z,y) \<in> r \<longrightarrow> y \<notin> Q))"
 then obtain Q where "x \<in> Q" and "\<forall>z\<in>Q. \<exists>y. (z,y) \<in> r \<and> y \<in> Q" by auto
 hence "\<forall>z. \<exists>y. z \<in> Q \<longrightarrow> (z,y) \<in> r \<and> y \<in> Q" by auto
 hence "\<exists>f. \<forall>x. x \<in> Q \<longrightarrow> (x,f x) \<in> r \<and> f x \<in> Q" by (rule choice)
 then obtain f where a: "\<forall>x. x \<in> Q \<longrightarrow> (x,f x) \<in> r \<and> f x \<in> Q" (is "\<forall>x. ?P x") by best
 let ?S = "\<lambda>i. T f x i"
 have "?S 0 = x" by simp
 have "\<forall>i. (?S i,?S(Suc i)) \<in> r \<and> ?S(Suc i) \<in> Q"
 proof
  fix i show "(?S i,?S(Suc i)) \<in> r \<and> ?S(Suc i) \<in> Q" by (induct i) (auto simp:`x \<in> Q` a)
 qed
 with `?S 0 = x` have "\<exists>S. S 0 = x \<and> (\<forall>i. (S i,S(Suc i)) \<in> r)" by fast
 with assms show False by auto
qed

hide const T

lemma minimal_imp_wf:
-- {*
 x \<in> Q \<longrightarrow> ... for every non-empty subset
 z \<in> Q \<longrightarrow> ... there exists a minimal element w.r.t. r
*}
 assumes "\<forall>Q x. x \<in> Q \<longrightarrow> (\<exists>z\<in>Q. \<forall>y. (z,y) \<in> r \<longrightarrow> y \<notin> Q)"
 shows "wf(r\<inverse>)"
proof (rule ccontr)
 assume "\<not> wf(r\<inverse>)"
 hence "\<exists>P. (\<forall>x. (\<forall>y. (x,y) \<in> r \<longrightarrow> P y) \<longrightarrow> P x) \<and> (\<exists>x. \<not> P x)" unfolding wf_def by simp
 then obtain P x where suc:"\<forall>x. (\<forall>y. (x,y) \<in> r \<longrightarrow> P y) \<longrightarrow> P x" and "\<not> P x" by auto
 let ?Q = "{x. \<not> P x}"
 from `\<not> P x` have "x \<in> ?Q" by simp
 from assms have "\<forall>x. x \<in> ?Q \<longrightarrow> (\<exists>z\<in>?Q. \<forall>y. (z,y) \<in> r \<longrightarrow> y \<notin> ?Q)" by (rule allE[where x = ?Q])
 with `x \<in> ?Q` obtain z where "z \<in> ?Q" and min:" \<forall>y. (z,y) \<in> r \<longrightarrow> y \<notin> ?Q" by best
 from `z \<in> ?Q` have "\<not> P z" by simp
 with suc obtain y where "(z,y) \<in> r" and "\<not> P y" by best
 hence "y \<in> ?Q" by simp
 with `(z,y) \<in> r` and min show False by simp
qed

lemmas SN_imp_wf = SN_imp_minimal[THEN minimal_imp_wf]

lemma SN_iff_wf: "SN r = wf(r\<inverse>)" by (auto simp: SN_imp_wf wf_imp_SN)

lemma Newman's_Lemma:
 fixes r :: "'a brel" assumes "SN r" and "WCR r" shows "CR r"
proof - {
 fix x :: 'a
 from `SN r` have "wf(r\<inverse>)" unfolding SN_iff_wf by simp
 hence "CR r x"
 proof induct
  case (less x)
  hence IH: "\<forall>y. (x,y) \<in> r \<longrightarrow> CR r y" by simp
  {
   fix y z assume "(x,y) \<in> r^*" and "(x,z) \<in> r^*"
   from `(x,y) \<in> r^*` obtain m where "(x,y) \<in> r^m" using rtrancl_imp_UN_rel_pow by best
   from `(x,z) \<in> r^*` obtain n where "(x,z) \<in> r^n" using rtrancl_imp_UN_rel_pow by best
   hence "(y,z) \<in> r^\<down>"
   proof (cases n)
    assume "(x,z) \<in> r^n" and "n = 0"
    hence "x = z" unfolding `n = 0` by simp
    with `(x,y) \<in> r^*` show ?thesis by auto
   next
    fix n' assume "(x,z) \<in> r^n" and "n = Suc n'"
    from `(x,y) \<in> r^m` show ?thesis
    proof (cases m)
     assume "(x,y) \<in> r^m" and "m = 0"
     hence "x = y" unfolding `m = 0` by simp
     with `(x,z) \<in> r^*` show ?thesis by auto
    next
     fix m' assume "(x,y) \<in> r^m" and "m = Suc m'"
     from `(x,z) \<in> r^n` have "(x,z) \<in> r^Suc n'" unfolding `n = Suc n'` by simp
     then obtain z' where "(x,z') \<in> r" and "(z',z) \<in> r^n'" using rel_pow_Suc_D2 by best
     hence "(z',z) \<in> r^*" using rel_pow_imp_rtrancl by auto
     from `(x,y) \<in> r^m` have "(x,y) \<in> r^Suc m'" unfolding `m = Suc m'` by simp
     then obtain y' where "(x,y') \<in> r" and "(y',y) \<in> r^m'" using rel_pow_Suc_D2 by best
     hence "(y',y) \<in> r^*" using rel_pow_imp_rtrancl by auto
     from `WCR r` have "WCR r x" by auto
     with `(x,y') \<in> r` and `(x,z') \<in> r` have "(y',z') \<in> r^\<down>" by auto
     then obtain u where "(y',u) \<in> r^*" and "(z',u) \<in> r^*" by best
     from IH and `(x,y') \<in> r` have "CR r y'" by simp
     with `(y',y) \<in> r^*` and `(y',u) \<in> r^*` have "(y,u) \<in> r^\<down>" by auto
     then obtain v where "(y,v) \<in> r^*" and "(u,v) \<in> r^*" by auto
     from IH and `(x,z') \<in> r` have "CR r z'" by simp
     from `(z',u) \<in> r^*` and `(u,v) \<in> r^*` have "(z',v) \<in> r^*" by simp
     with `CR r z'` and `(z',z) \<in> r^*` have "(v,z) \<in> r^\<down>" by auto
     then obtain w where "(v,w) \<in> r^*" and "(z,w) \<in> r^*" by auto
     from `(y,v) \<in> r^*` and `(v,w) \<in> r^*` have "(y,w) \<in> r^*" by auto
     with `(z,w) \<in> r^*` show ?thesis by auto
    qed
   qed
  } thus ?case by auto
 qed
} thus ?thesis by auto
qed

lemma all_reducts_SN_imp_SN:
 fixes r :: "'a brel" assumes "(\<And>y. (x,y) \<in> r \<Longrightarrow> SN r y)" shows "SN r x"
proof (rule ccontr)
 assume "\<not> SN r x"
 then obtain S where "S 0 = x" and inf:"\<forall>i. (S i,S(Suc i)) \<in> r" by auto
 let ?S = "\<lambda>i. S(Suc i)"
 from inf have "(S 0,?S 0) \<in> r" by simp
 with assms have "SN r (?S 0)" unfolding `S 0 = x` by simp
 hence a: "\<not>(\<exists>S. S 0 = (?S 0) \<and> (\<forall>i. (S i,S(Suc i)) \<in> r))" by auto
 from inf have "\<forall>i. (?S i,?S(Suc i)) \<in> r" by simp
 hence "\<exists>S. S 0 = ?S 0 \<and> (\<forall>i. (S i,S(Suc i)) \<in> r)" by auto
 with a show False by simp
qed

definition "restrict" :: "'a brel \<Rightarrow> 'a set \<Rightarrow> 'a brel"
where "restrict r S \<equiv> {(x,y). x \<in> S \<and> y \<in> S \<and> (x,y) \<in> r}"

lemma SN_imp_restrict_SN_elt:
 fixes r :: "'a brel" and x :: 'a assumes "SN r x" shows "SN(restrict r S) x" (is "SN ?r x")
proof (rule ccontr)
 assume "\<not> SN ?r x"
 hence "\<exists>S. S 0 = x \<and> (\<forall>i. (S i,S(Suc i)) \<in> ?r)" by auto
 hence "\<exists>S. S 0 = x \<and> (\<forall>i. (S i,S(Suc i)) \<in> r)" unfolding restrict_def by auto
 with `SN r x` show False by auto
qed

lemma SN_imp_restrict_SN: "SN(r::'a brel) \<Longrightarrow> SN(restrict r S)"
using SN_imp_restrict_SN_elt unfolding SN_ars_def by best

lemma restrict_rtrancl: "(restrict r S)^* \<subseteq> r^*" (is "?r^* \<subseteq> r^*")
proof - {
 fix x y assume "(x,y) \<in> ?r^*" hence "(x,y) \<in> r^*" unfolding restrict_def by induct auto
} thus ?thesis by auto
qed

lemma WCR_SN_elt_imp_CR_elt:
 fixes r :: "'a brel" and x :: 'a assumes "WCR r" and "SN r x" shows "CR r x"
proof -
 let ?S = "{y. (x,y) \<in> r^*}"
 let ?r = "restrict r ?S"
 have "\<forall>x. (SN::'a brel \<Rightarrow> 'a set) ?r x"
 proof
  fix y have "y \<notin> ?S \<or> y \<in> ?S" by simp 
  thus "SN ?r y"
  proof
   assume "y \<notin> ?S" thus ?thesis unfolding restrict_def by auto
  next
   assume "y \<in> ?S"
   hence "(x,y) \<in> r^*" by simp
   with `SN r x` have "SN r y" by (rule SN_imp_SN_reduct)
   thus ?thesis by (rule SN_imp_restrict_SN_elt)
  qed
 qed
 hence "SN ?r" by auto
 {
  fix x y assume "(x,y) \<in> r^*" and "x \<in> ?S" and "y \<in> ?S"
  then obtain n where "(x,y) \<in> r^n" and "x \<in> ?S" and "y \<in> ?S" using rtrancl_imp_UN_rel_pow by best
  hence "(x,y) \<in> ?r^*"
  proof (induct n arbitrary: x y)
   case 0 thus ?case by simp
  next
   case (Suc n)
   from `(x,y) \<in> r^Suc n` obtain x' where "(x,x') \<in> r" and "(x',y) \<in> r^n" using rel_pow_Suc_D2 by best
   hence "(x,x') \<in> r^*" by simp
   with `x \<in> ?S` have "x' \<in> ?S" by auto
   with Suc and `(x',y) \<in> r^n` have "(x',y) \<in> ?r^*" by simp
   from `(x,x') \<in> r` and `x \<in> ?S` and `x' \<in> ?S` have "(x,x') \<in> ?r" unfolding restrict_def by simp
   with `(x',y) \<in> ?r^*` show ?case by simp
  qed
 }
 hence a:"\<forall>x y. (x,y) \<in> r^* \<and> x \<in> ?S \<and> y \<in> ?S \<longrightarrow> (x,y) \<in> ?r^*" by simp
 {
  fix x' y z assume "(x',y) \<in> ?r" and "(x',z) \<in> ?r"
  hence "x' \<in> ?S" and "y \<in> ?S" and "z \<in> ?S" and "(x',y) \<in> r" and "(x',z) \<in> r"
   unfolding restrict_def by auto
  with `WCR r` have "(y,z) \<in> r^\<down>" by force
  then obtain u where "(y,u) \<in> r^*" and "(z,u) \<in> r^*" by auto
  from `x' \<in> ?S` have "(x,x') \<in> r^*" by simp
  from `(x',y) \<in> r` have "(x',y) \<in> r^*" by auto
  with `(y,u) \<in> r^*` have "(x',u) \<in> r^*" by auto
  with `(x,x') \<in> r^*` have "(x,u) \<in> r^*" by simp
  hence "u \<in> ?S" by simp
  from `y \<in> ?S` and `u \<in> ?S` and `(y,u) \<in> r^*` have "(y,u) \<in> ?r^*" using a by auto
  from `z \<in> ?S` and `u \<in> ?S` and `(z,u) \<in> r^*` have "(z,u) \<in> ?r^*" using a by auto
  with `(y,u) \<in> ?r^*` have "(y,z) \<in> ?r^\<down>" by auto
 }
 hence "WCR ?r" by auto
 have "CR ?r" using Newman's_Lemma[OF `SN ?r` `WCR ?r`] by simp
 {
  fix y z assume "(x,y) \<in> r^*" and "(x,z) \<in> r^*"
  hence "y \<in> ?S" and "z \<in> ?S" by auto
  have "x \<in> ?S" by simp
  from a and `(x,y) \<in> r^*` and `x \<in> ?S` and `y \<in> ?S` have "(x,y) \<in> ?r^*" by simp
  from a and `(x,z) \<in> r^*` and `x \<in> ?S` and `z \<in> ?S` have "(x,z) \<in> ?r^*" by simp
  with `CR ?r` and `(x,y) \<in> ?r^*` have "(y,z) \<in> ?r^\<down>" by force
  then obtain u where "(y,u) \<in> ?r^*" and "(z,u) \<in> ?r^*" by best
  hence "(y,u) \<in> r^*" and "(z,u) \<in> r^*" using restrict_rtrancl by auto
  hence "(y,z) \<in> r^\<down>" by auto
 }
 thus ?thesis by auto
qed

lemma rtrancl_imp_seq:
 assumes "(x,y) \<in> r^*"
 shows "\<exists>S n. S 0 = x \<and> S n = y \<and> (\<forall>i<n. (S i,S(Suc i)) \<in> r)"
using assms proof (induct)
 case base thus ?case by auto
next
 case (step y z)
 then obtain S n where "S 0 = x" and "S n = y" and seq: "\<forall>i<n. (S i,S(Suc i)) \<in> r" by auto
 let ?m = "Suc n"
 let ?S = "\<lambda>i. if i = ?m then z else S i"
 have "?S ?m = z" by simp
 from `S 0 = x` have "?S 0 = x" by simp
 from seq have seq': "\<forall>i<n. (?S i,?S(Suc i)) \<in> r" by auto
 with `S n = y` and `(y,z) \<in> r` have "\<forall>i<?m. (?S i,?S(Suc i)) \<in> r" by auto
 with `?S 0 = x` and `?S ?m = z` show ?case by best
qed

lemma SN_imp_normalizable: fixes r :: "'a brel" assumes "SN r x" shows "\<exists>y. (x,y) \<in> r^!"
proof (rule ccontr)
 assume "\<not>(\<exists>y. (x,y) \<in> r^!)"
 hence A: "\<forall>y. (x,y) \<in> r^* \<longrightarrow> y \<notin> NF r" by auto
 hence "x \<notin> NF r" by auto
 let ?Q = "{z. (x,z) \<in> r^* \<and> z \<notin> NF r}"
 have "x \<in> ?Q" using `x \<notin> NF r` by simp
 have "\<forall>z\<in>?Q. \<exists>y. (z,y) \<in> r \<and> y \<in> ?Q"
 proof
  fix z assume "z \<in> ?Q"
  hence "(x,z) \<in> r^*" and "z \<notin> NF r" by auto
  then obtain u where "(z,u) \<in> r" by auto
  with `(x,z) \<in> r^*` have "(x,u) \<in> r^*" by simp
  with A have "u \<notin> NF r" by simp
  with `(z,u) \<in> r` and `(x,u) \<in> r^*`
  show "\<exists>y. (z,y) \<in> r \<and> y \<in> ?Q" by auto
 qed
 with `x \<in> ?Q` have "x \<in> ?Q \<and> (\<forall>z\<in>?Q. \<exists>y. (z,y) \<in> r \<and> y \<in> ?Q)" by auto
 hence "\<exists>Q. x \<in> Q \<and> (\<forall>z\<in>Q. \<exists>y. (z,y) \<in> r \<and> y \<in> Q)" by (rule exI[where x = "?Q"])
 hence "\<not>(\<forall>Q. x \<in> Q \<longrightarrow> (\<exists>z\<in>Q. \<forall>y. (z,y) \<in> r \<longrightarrow> y \<notin> Q))" by simp
 with SN_elt_imp_elt_minimal[of r x] have "\<not>(SN r x)" by auto
 with assms show False by simp
qed

subsection {* Commutation *}

definition commute :: "'a brel \<Rightarrow> 'a brel \<Rightarrow> bool"
where "commute r s \<equiv> (s^* O (r\<inverse>)^*) \<subseteq> ((r\<inverse>)^* O s^*)"

lemma CR_iff_self_commute: "CR r = commute r r"
unfolding commute_def CR_iff_meetable_subset_joinable meetable_def joinable_def by simp

lemma rtrancl_imp_rtrancl_UN: 
 assumes "(x,y) \<in> r^*" and "r \<in> I"
 shows "(x,y) \<in> (\<Union>r\<in>I. r)^*" (is "(x,y) \<in> ?r^*")
using assms proof induct
 case base thus ?case by simp
next
 case (step y z)
 hence "(x,y) \<in> ?r^*" by simp
 from `(y,z) \<in> r` and `r \<in> I` have "(y,z) \<in> ?r^*" by auto
 with `(x,y) \<in> ?r^*` show ?case by auto
qed

(*
lemma
 fixes I :: "('a brel)set"
 assumes "\<And>r s.\<lbrakk>r \<in> I; s \<in> I\<rbrakk> \<Longrightarrow> commute r s"
 shows "(x,y) \<in> (\<Union>r\<in>I. r)^n \<Longrightarrow> (x,z) \<in> (\<Union>r\<in>I. r) \<Longrightarrow> (y,z) \<in> (\<Union>r\<in>I. r)^\<down>" (is "_ \<Longrightarrow> _ \<Longrightarrow> _ \<in> ?r^\<down>")
proof -
 assume "(x,y) \<in> ?r^n" and "(x,z) \<in> ?r"
 then obtain s where "(x,z) \<in> s" and "s \<in> I" by best
 hence "(x,z) \<in> s^*" by auto
 hence "(x,z) \<in> ?r^*" by auto
 with `(x,y) \<in> ?r^n` and `(x,z) \<in> ?r` show "(y,z) \<in> ?r^\<down>"
 proof (induct n arbitrary: x y z)
  case 0 thus ?case by auto
 next
  case (Suc n)
  from `(x,y) \<in> ?r^Suc n` obtain y' where "(x,y') \<in> ?r" and "(y',y) \<in> ?r^n" using rel_pow_Suc_D2 by best
*)

definition quasi_commute :: "'a brel \<Rightarrow> 'a brel \<Rightarrow> bool"
where "quasi_commute r s \<equiv> (r O s) \<subseteq> (r \<union> s)^* O r"

lemma rtrancl_union_subset_rtrancl_union_trancl: "(r \<union> s^+)^* = (r \<union> s)^*"
proof
show "(r \<union> s^+)^* \<subseteq> (r \<union> s)^*"
 proof (rule subsetI2)
  fix x y assume "(x,y) \<in> (r \<union> s^+)^*"
  thus "(x,y) \<in> (r \<union> s)^*"
  proof (induct)
   case base thus ?case by auto
  next
   case (step y z)
   hence "(y,z) \<in> r \<or> (y,z) \<in> s^+" by auto
   hence "(y,z) \<in> (r \<union> s)^*"
   proof
    assume "(y,z) \<in> r" thus ?thesis by auto
   next
    assume "(y,z) \<in> s^+"
    hence "(y,z) \<in> s^*" by auto
    hence "(y,z) \<in> r^* \<union> s^*" by auto
    thus ?thesis using rtrancl_Un_subset by auto
   qed
   with `(x,y) \<in> (r \<union> s)^*` show ?case by simp
  qed
 qed
next
 show "(r \<union> s)^* \<subseteq> (r \<union> s^+)^*"
 proof (rule subsetI2)
  fix x y assume "(x,y) \<in> (r \<union> s)^*"
  thus "(x,y) \<in> (r \<union> s^+)^*"
  proof (induct)
   case base thus ?case by auto
  next
   case (step y z)
   hence "(y,z) \<in> (r \<union> s^+)^*" by auto
   with `(x,y) \<in> (r \<union> s^+)^*` show ?case by auto
  qed
 qed
qed

lemma qc_imp_qc_trancl: assumes "quasi_commute r s" shows "quasi_commute r (s^+)"
unfolding quasi_commute_def proof (rule subsetI2)
 fix x z assume "(x,z) \<in> r O s^+"
 then obtain y where "(x,y) \<in> s^+" and "(y,z) \<in> r" by best
 thus "(x,z) \<in> (r \<union> s^+)^* O r"
 proof (induct arbitrary: z)
  case (base y)
  hence "(x,z) \<in> (r O s)" by auto
  with assms have "(x,z) \<in> (r \<union> s)^* O r" unfolding quasi_commute_def by auto
  thus ?case using rtrancl_union_subset_rtrancl_union_trancl by auto
 next
  case (step a b)
  hence "(a,z) \<in> (r O s)" by auto
  with assms have "(a,z) \<in> (r \<union> s)^* O r" unfolding quasi_commute_def by auto
  then obtain u where "(a,u) \<in> r" and "(u,z) \<in> (r \<union> s)^*" by best
  hence "(u,z) \<in> (r \<union> s^+)^*" using rtrancl_union_subset_rtrancl_union_trancl by auto
  from `(a,u) \<in> r` and step have "(x,u) \<in> (r \<union> s^+)^* O r" by auto
  then obtain v where "(x,v) \<in> r" and "(v,u) \<in> (r \<union> s^+)^*" by best
  with `(u,z) \<in> (r \<union> s^+)^*` have "(v,z) \<in> (r \<union> s^+)^*" by auto
  with `(x,v) \<in> r` show ?case by auto
 qed
qed

lemma not_SN_reduct_imp_not_SN:
 fixes r :: "'a brel"
 assumes "\<not> SN r y" and "(x,y) \<in> r^*"
 shows "\<not> SN r x"
using SN_imp_SN_reduct[where r = r and x = x and y = y] assms by auto


lemma elt_of_nontermin_seq_is_nontermin:
 assumes "\<And>i. (S i,S(Suc i)) \<in> r" shows "\<And>i. \<not> SN r (S i)"
proof -
 fix i
 let ?S = "\<lambda>j. S(i + j)"
 have "?S 0 = S i" by simp
 have "\<forall>i. (?S i,?S(Suc i)) \<in> r" using assms by auto
 with `?S 0 = S i` have "?S 0 = S i \<and> (\<forall>i. (?S i,?S(Suc i)) \<in> r)" by auto
 hence "\<exists>T. T 0 = S i \<and> (\<forall>i. (T i,T(Suc i)) \<in> r)" by best
 thus "\<not> SN r (S i)" unfolding SN_elt_def by auto
qed

lemma quasi_commute_imp_SN:
 fixes r :: "'a brel"
 assumes "SN r" and "SN s" and "quasi_commute r s"
 shows "SN(r \<union> s)"
proof -
 have "quasi_commute r (s^+)" by (rule qc_imp_qc_trancl[OF `quasi_commute r s`])
 let ?B = "{a. \<not> (SN::'a brel \<Rightarrow> 'a \<Rightarrow> bool) (r \<union> s) a}"
 {
  assume "\<not> SN(r \<union> s)"
  then obtain a where "a \<in> ?B" unfolding SN_ars_def by best
  from `SN r` have "\<forall>Q x. x \<in> Q \<longrightarrow> (\<exists>z\<in>Q. \<forall>y. (z,y) \<in> r \<longrightarrow> y \<notin> Q)" by (rule SN_imp_minimal)
  hence "\<forall>x. x \<in> ?B \<longrightarrow> (\<exists>z\<in>?B. \<forall>y. (z,y) \<in> r \<longrightarrow> y \<notin> ?B)" by (rule spec[where x = ?B])
  with `a \<in> ?B` obtain b where "b \<in> ?B" and min: "\<forall>y. (b,y) \<in> r \<longrightarrow> y \<notin> ?B" by auto
  from `b \<in> ?B` obtain S where "S 0 = b" and
  seq: "\<forall>i. (S i,S(Suc i)) \<in> r \<union> s" unfolding SN_elt_def by auto
  let ?S = "\<lambda>i. S(Suc i)"
  have "?S 0 = S 1" by simp
  from seq have "\<forall>i. (?S i,?S(Suc i)) \<in> r \<union> s" by auto
  with `?S 0 = S 1` have "\<not> SN (r \<union> s) (S 1)" unfolding SN_elt_def by auto
  from `S 0 = b` and seq have "(b,S 1) \<in> r \<union> s" by auto
  with min and `\<not> SN (r \<union> s) (S 1)` have "(b,S 1) \<in> s" by auto
  let ?i = "LEAST i. (S i,S(Suc i)) \<notin> s"
  {
   assume "\<forall>i. (S i,S(Suc i)) \<in> s"
   with `S 0 = b` have "\<not> SN s b" unfolding SN_elt_def by auto
   with `SN s` have False unfolding SN_ars_def by auto
  }
  hence ex: "\<exists>i. (S i,S(Suc i)) \<notin> s" by auto
  hence "(S ?i,S(Suc ?i)) \<notin> s" by (rule LeastI_ex)
  with seq have "(S ?i,S(Suc ?i)) \<in> r" by auto
  have ini: "\<forall>i<?i. (S i,S(Suc i)) \<in> s" using not_less_Least by auto
  {
   fix i assume "i < ?i" hence "(b,S(Suc i)) \<in> s^+"
   proof (induct i)
    case 0 thus ?case using `(b,S 1) \<in> s` and `S 0 = b` by auto
   next
    case (Suc k)
    hence "(b,S(Suc k)) \<in> s^+" and "Suc k < ?i" by auto
    with `\<forall>i<?i. (S i,S(Suc i)) \<in> s` have "(S(Suc k),S(Suc(Suc k))) \<in> s" by fast
    with `(b,S(Suc k)) \<in> s^+` show ?case by auto
   qed
  }
  hence pref: "\<forall>i<?i. (b,S(Suc i)) \<in> s^+" by auto
  from `(b,S 1) \<in> s` and `S 0 = b` have "(S 0,S(Suc 0)) \<in> s" by auto
  {
   assume "?i = 0"
   from ex have "(S ?i,S(Suc ?i)) \<notin> s" by (rule LeastI_ex)
   with `(S 0,S(Suc 0)) \<in> s` have False unfolding `?i = 0` by simp
  }
  hence "0 < ?i" by auto
  then obtain j where "?i =  Suc j" unfolding gr0_conv_Suc by best
  with ini have "(S(?i-Suc 0),S(Suc(?i-Suc 0))) \<in> s" by auto
  with pref have "(b,S(Suc j)) \<in> s^+" unfolding `?i = Suc j` by auto
  hence "(b,S ?i) \<in> s^+" unfolding `?i = Suc j` by auto
  with `(S ?i,S(Suc ?i)) \<in> r` have "(b,S(Suc ?i)) \<in> (r O s^+)" by auto
  with `quasi_commute r (s^+)` have "(b,S(Suc ?i)) \<in> (r \<union> s^+)^* O r" unfolding quasi_commute_def by auto
  then obtain c where "(b,c) \<in> r" and "(c,S(Suc ?i)) \<in> (r \<union> s^+)^*" by best
  from `(b,c) \<in> r` have "(b,c) \<in> (r \<union> s)^*" by auto
  from elt_of_nontermin_seq_is_nontermin[where r = "r \<union> s" and S = S]
    and seq have "\<not> SN (r \<union> s) (S(Suc ?i))" by auto
  from `(c,S(Suc ?i)) \<in> (r \<union> s^+)^*` have "(c,S(Suc ?i)) \<in> (r \<union> s)^*" unfolding rtrancl_union_subset_rtrancl_union_trancl by auto
  with not_SN_reduct_imp_not_SN[where r = "r \<union> s" and x = c] and `\<not> SN (r \<union> s) (S(Suc ?i))` have "\<not> SN (r \<union> s) c" by auto
  hence "c \<in> ?B" by simp
  with `(b,c) \<in> r` and min have False by auto
 }
 thus ?thesis by auto
qed


lemma non_strict_into_strict : assumes compat: "S O NS \<subseteq> S"
  and steps: "(s,t) \<in> S O (NS^*)"
  shows "(s,t) \<in> S"
using steps
proof auto
  fix u
  assume "(s, u) \<in> NS^*" and "(u,t) \<in> S"
  thus ?thesis 
  proof (induct rule:rtrancl.induct, simp)
    case (rtrancl_into_rtrancl a b c)
    with compat show ?case by auto
  qed
qed


lemma non_strict_ending : assumes steps: "\<forall> i. (f i, f (Suc i)) \<in> NS \<union> S"
  and compat: "S O NS \<subseteq> S"
  and SN: "SN S (f 0)"
  shows "\<exists> j. \<forall> i \<ge> j. (f i, f (Suc i)) \<in> NS - S"
proof (rule ccontr, auto)
  assume "\<forall>j. \<exists> i \<ge> j. (f i, f (Suc i)) \<in>  NS \<longrightarrow> (f i, f (Suc i)) \<in> S"
  with steps have Ssteps: "\<forall>j. \<exists> i \<ge> j. (f i, f (Suc i)) \<in> S" by blast
  let ?g = "\<lambda> (i :: nat). LEAST j. card {k. k \<le> j \<and> (f k, f (Suc k)) \<in> S} = Suc i "
  have gWD: "\<forall> i. \<exists> j. card {k. k \<le> j \<and> (f k, f (Suc k)) \<in> S} = (Suc i)" 
  proof 
    fix i
    show "\<exists> j. card {k. k \<le> j \<and> (f k, f (Suc k)) \<in> S} = Suc i"
    proof (induct i)
      case 0
      let ?j = "LEAST j. (f j, f (Suc j)) \<in> S"
      have "{k. k \<le> ?j \<and> (f k, f (Suc k)) \<in> S} = {?j}" (is "?S = _") 
      proof 
	from Ssteps obtain k where "(f k, f (Suc k)) \<in> S" by auto
	hence "(f ?j, f (Suc ?j)) \<in> S" by (rule LeastI)
	thus "{?j} \<subseteq> ?S" by auto
      next
	show "?S \<subseteq> {?j}"
	proof auto
	  fix k 
	  assume "(f k, f (Suc k)) \<in> S"
	  hence one: "?j \<le> k" by (rule Least_le)
	  assume "k \<le> ?j"
	  with one show "k = ?j" by simp
	qed
      qed
      from this have "card {k. k \<le> ?j \<and> (f k, f (Suc k)) \<in> S} = Suc 0" by simp
      thus ?case by blast
    next
      case (Suc i)
      from this obtain j where cardSI: "card {k. k \<le> j \<and> (f k, f (Suc k)) \<in> S} = Suc i" by auto
      let ?j = "LEAST jj. jj > j \<and> (f jj, f (Suc jj)) \<in> S"
      from Ssteps obtain jjj where "jjj \<ge> (Suc j)" and "(f jjj, f (Suc jjj)) \<in> S" by auto
      hence "jjj > j \<and> (f jjj, f (Suc jjj)) \<in> S" by auto
      hence jSat: "?j > j \<and> (f ?j, f (Suc ?j)) \<in> S" by (rule LeastI)
      hence jSat2: "?j \<le> ?j \<and> (f ?j, f (Suc ?j)) \<in> S" by auto
      hence jj1: "?j \<in> {k. k \<le> ?j \<and> (f k, f (Suc k)) \<in> S}" (is "_ \<in> ?L")  by auto
      from jSat have jj2: "?j \<notin> {k. k \<le> j \<and> (f k, f (Suc k)) \<in> S}" (is "_ \<notin> ?S") by auto 
      with jSat have jj3: "?S \<subseteq> ?L" by auto
      with jj1 have jj4: "insert ?j ?S \<subseteq> ?L" by simp
      have jj5: "?L \<subseteq> insert ?j ?S" 
      proof
	fix k
	assume "k \<in> ?L"
	hence kj: "k \<le> ?j" and kS: "(f k, f (Suc k)) \<in> S" by auto
	show "k \<in> insert ?j ?S"
	proof (cases "k \<le> j \<or> k = ?j")
	  case True with kS show ?thesis by auto
	next
	  case False
	  with kj have "j < k \<and> k < ?j" by simp
	  with kS not_less_Least have False by blast
	  thus ?thesis ..
	qed
      qed	  
      with jj4 have jj6 : "insert ?j ?S = ?L" by auto
      
      have fin1: "finite {k. (k :: nat) \<le> (j :: nat)}" proof (induct j, auto)
	fix j
	assume "finite {k. k \<le> (j :: nat)}"
	hence A: "finite (insert (Suc j) {k. k \<le> j})" ..
	have "insert (Suc j) {k. k \<le> j} = {k. k \<le> Suc j}" by auto
	with A show "finite {k. k \<le> (Suc j)}"  by force
      qed
      have "?S \<subseteq> {k. k \<le> j}" by auto 
      from this and fin1 have fin2: "finite ?S" by (rule finite_subset)
      from this and jj2 have "card (insert ?j ?S) = Suc (card ?S)" by (rule card_insert_disjoint)
      with jj6 and cardSI have "card ?L = Suc (Suc i)" by simp
      thus ?case ..
    qed
  qed

  have endSteps: "\<forall> i. (f (?g i), f (Suc (?g i))) \<in> S"
  proof
    fix i
    { 
      assume nE: "(f (?g i), f (Suc (?g i))) \<notin> S"
      from gWD obtain j where "card {k. k \<le> j \<and> (f k, f (Suc k)) \<in> S} = Suc i" by auto
      hence card: "card {k. k \<le> ?g i \<and> (f k, f (Suc k)) \<in> S} = Suc i" by (rule LeastI)
      with nE have "?g i \<notin> {k. k \<le> ?g i \<and> (f k, f (Suc k)) \<in> S}" by auto
      with card have cardII: "card ({k. k \<le> ?g i \<and> (f k, f (Suc k)) \<in> S} - {?g i}) = Suc i" by simp
      have False
      proof (cases "?g i")
	case 0
	hence "{k. k \<le> ?g i \<and> (f k, f (Suc k)) \<in> S} - {?g i} = {}" by auto
	with cardII show False by auto
      next
	case (Suc n)
	hence "n < ?g i" by simp
	have "{k. (k :: nat) \<le> ?g i \<and> (f k, f (Suc k)) \<in> S} - {?g i} = {k. k \<le> n \<and> (f k, f (Suc k)) \<in> S}" using Suc nE 
	proof (auto)
	  fix x
	  assume "(f (Suc n), f (Suc (Suc n))) \<notin> S" and "x \<le> Suc n" and "(f x, f (Suc x)) \<in> S"
	  thus "x \<le> n" by (cases "x = Suc n", auto)
	qed
	with cardII have "card {k. k \<le> n \<and> (f k, f (Suc k)) \<in> S} = Suc i" by auto
	with `n < ?g i` not_less_Least show False by blast
      qed      
    }
    thus "(f (?g i), f (Suc (?g i))) \<in> S" by auto
  qed

  have setMono: "\<And> i j. i \<le> j \<Longrightarrow> {k. k \<le> i \<and> (f k, f (Suc k)) \<in> S} \<subseteq> {k. k \<le> j \<and> (f k, f (Suc k)) \<in> S}" by auto
  have gMono: "\<And> i. ?g i < ?g (Suc i)" 
  proof -
    fix i
    let ?S = "{k. k \<le> ?g i \<and> (f k, f (Suc k)) \<in> S}"
    let ?L = "{k. k \<le> ?g (Suc i) \<and> (f k, f (Suc k)) \<in> S}"
    from gWD obtain jL where cardjL: "card {k. k \<le> jL \<and> (f k, f (Suc k)) \<in> S} = Suc (Suc i)" by auto
    hence cardL: "card ?L = Suc (Suc i)" by (rule LeastI)

    from gWD obtain jS where cardjS: "card {k. k \<le> jS \<and> (f k, f (Suc k)) \<in> S} = Suc i" by auto
    hence cardS: "card ?S = Suc i" by (rule LeastI)

    show "?g i < ?g (Suc i)" 
    proof (rule ccontr)
      assume "\<not> ?g i < ?g (Suc i)"
      hence "?g (Suc i) \<le> ?g i" by simp
      hence subset: "?L \<subseteq> ?S" by (rule setMono)

      from cardS have "\<not> card ?S = 0" by simp
      hence finL: "finite ?S" by (auto simp: card_eq_0_iff)
      from this and subset have "card ?L \<le> card ?S" by (rule card_mono)
      with cardL cardS show False by simp
    qed
  qed
    
  have interSteps: "\<And> i l. \<lbrakk>?g i < l; l < ?g (Suc i)\<rbrakk> \<Longrightarrow> (f l, f (Suc l)) \<notin> S" 
  proof
    fix i l :: nat
    let ?S = "{k. k \<le> ?g i \<and> (f k, f (Suc k)) \<in> S}"
    let ?M = "{k. k \<le> l \<and> (f k, f (Suc k)) \<in> S}"
    let ?L = "{k. k \<le> ?g (Suc i) \<and> (f k, f (Suc k)) \<in> S}"
    assume "l < ?g (Suc i)"
    hence "l \<le> ?g (Suc i)" by auto
    hence subs: "?M \<subseteq> ?L" by (rule setMono)
    from `l < ?g (Suc i)` have siM: "?g (Suc i) \<notin> ?M" by auto
    from endSteps have "?g (Suc i) \<in> ?L" by blast
    with subs siM have ml: "?M \<subset> ?L" by auto
    assume "?g i < l"
    hence "?g i \<le> l" by simp
    hence subs2: "?S \<subseteq> ?M" by (rule setMono)
    from `?g i < l` have niS: "l \<notin> ?S" by auto
    assume "(f l, f (Suc l)) \<in> S"
    hence "l \<in> ?M" by auto
    with niS subs2 have sm: "?S \<subset> ?M" by auto    
    
    from gWD obtain j2 where "card {k. k \<le> j2 \<and> (f k, f (Suc k)) \<in> S} = Suc (Suc i)" by auto
    hence cardL: "card ?L = Suc (Suc i)" by (rule LeastI)
    hence "\<not> card ?L = 0" by simp
    hence finL: "finite ?L" by (auto simp: card_eq_0_iff)
    from this and ml have "card ?M < card ?L" by (rule psubset_card_mono)
    with cardL have oneDir: "card ?M \<le> Suc i" by simp

    from gWD obtain j where "card {k. k \<le> j \<and> (f k, f (Suc k)) \<in> S} = Suc i" by auto
    hence cardS: "card ?S = Suc i" by (rule LeastI)
    from subs finL have "finite ?M" by (rule finite_subset)
    from this and sm have "card ?S < card ?M" by (rule psubset_card_mono)
    with cardS oneDir show False by simp
  qed

  have allStrict: "\<forall> i. (f (Suc (?g i)), f (Suc (?g (Suc i)))) \<in> S"
  proof 
    fix i
    let ?n = "?g (Suc i) - Suc (?g i)" 
    from gMono have "?g i < ?g (Suc i)" by auto 
    hence n: "Suc (?g i) + ?n = ?g (Suc i)" by simp
    have "\<forall> l. l \<le> ?n \<longrightarrow> (f (Suc (?g i)), f (Suc (?g i) + l)) \<in> NS^*"
    proof (rule allI, rule impI)
      fix l
      assume "l \<le> ?n"
      thus "(f (Suc (?g i)), f (Suc (?g i) + l)) \<in> NS^*"
      proof (induct l)
	case 0
	thus ?case by simp
      next
	case (Suc l)	
	show ?case 
	proof
	  from Suc show "(f (Suc (?g i)), f (Suc (?g i) + l)) \<in> NS^*" by auto
	next
	  from Suc n have "?g i < Suc (?g i) + l" (is "_ < ?z") and "Suc (?g i) + l < ?g (Suc i)" by auto 
	  hence "(f ?z, f (Suc ?z)) \<notin> S" by (rule interSteps)
	  with steps have "(f ?z, f (Suc ?z)) \<in> NS" by auto
	  thus "(f ?z, f (Suc (?g i) + (Suc l))) \<in> NS" by simp
	qed
      qed
    qed 
    
    with n have partI: "(f (Suc (?g i)), f (?g (Suc i))) \<in> NS^*" by auto
    from endSteps have "(f (?g (Suc i)), f (Suc (?g (Suc i)))) \<in> S" by auto
    with partI have "(f (Suc (?g i)), f (Suc (?g (Suc i)))) \<in> S O NS^*" by auto
    with compat show "(f (Suc (?g i)), f (Suc (?g (Suc i)))) \<in> S" by (rule non_strict_into_strict)    
  qed

  have initialPart: "(f 0, f (Suc (?g 0))) \<in> S"
  proof -
    from endSteps have strict: "(f (?g 0), f (Suc (?g 0))) \<in> S" by simp

    have befSteps: "\<And> l. \<lbrakk>l < ?g 0\<rbrakk> \<Longrightarrow> (f l, f (Suc l)) \<notin> S" 
    proof 
      fix l :: nat
      let ?M = "{k. k \<le> l \<and> (f k, f (Suc k)) \<in> S}"
      let ?L = "{k. k \<le> ?g 0 \<and> (f k, f (Suc k)) \<in> S}"
      assume "l < ?g 0"
      hence "l \<le> ?g 0" by auto
      hence subs: "?M \<subseteq> ?L" by (rule setMono)
      from `l < ?g 0` have siM: "?g 0 \<notin> ?M" by auto
      from endSteps have "?g 0 \<in> ?L" by blast
      with subs siM have ml: "?M \<subset> ?L" by auto
    
      from gWD obtain j2 where "card {k. k \<le> j2 \<and> (f k, f (Suc k)) \<in> S} = Suc 0" by auto
      hence cardL: "card ?L = Suc 0" by (rule LeastI)
      hence "\<not> card ?L = 0" by simp
      hence finL: "finite ?L" by (auto simp: card_eq_0_iff)
      from this and ml have "card ?M < card ?L" by (rule psubset_card_mono)
      with cardL have oneDir: "card ?M = 0" by simp

      from subs finL have "finite ?M" by (rule finite_subset)
      with oneDir have empty: "?M = {}" by auto
      assume "(f l, f (Suc l)) \<in> S"
      with empty `l < ?g 0` show False  by auto
    qed
    
    have "\<forall> l. l < ?g 0 \<longrightarrow> (f 0, f (Suc l)) \<in> NS^*"
    proof (rule allI, rule impI)
      fix l
      assume "l < ?g 0" 
      thus "(f 0, f (Suc l)) \<in> NS^*" 
      proof (induct l)
	case 0
	with befSteps have "(f 0, f (Suc 0)) \<notin> S" by blast
	with steps show ?case by blast
      next
	case (Suc l)
	hence "l < ?g 0" by simp
	with Suc have init: "(f 0, f (Suc l)) \<in> NS^*" by blast
	from Suc and befSteps have "(f (Suc l), f (Suc (Suc l))) \<notin> S" by blast
	with steps have "(f (Suc l), f (Suc (Suc l))) \<in> NS" by blast
	with init show ?case by auto
      qed
    qed
    hence "(f 0, f (?g 0)) \<in> NS^*" by (cases "?g 0", auto)
    
    from this and strict have "(f 0, f (Suc (?g 0))) \<in> S O (NS^*)" by auto
    with compat show ?thesis  by (rule non_strict_into_strict)
  qed  

  from this and allStrict obtain g where one: "(f 0, g 0) \<in> S" and two: "\<forall> i. (g i, g (Suc i)) \<in> S" by auto
  let ?h = "\<lambda> x. (case x of 0 \<Rightarrow> f 0 | Suc n \<Rightarrow> g n)"
  have hOne: "\<forall> i. (?h i, ?h (Suc i)) \<in> S" 
  proof (rule allI)
    fix i
    show "(?h i, ?h (Suc i)) \<in> S" by (cases i, auto simp: one two)
  qed
  have hTwo: "SN S (?h 0)" using SN by simp
  from hOne hTwo obtain h where "\<forall> i. (h i, h (Suc i)) \<in> S" and "SN S (h 0)" by blast
  with SN show False by (unfold SN_elt_def, auto)     
qed


end